import os
import datetime
from timeit import default_timer as timer
from typing import Callable
import cProfile
import pstats
import io
import numpy as np
import random
import inspect
import sys
import time
import pickle
from itertools import combinations
import shutil
import math
import psutil
import platform
import glob
from pathlib import Path
from uuid import getnode as get_mac
from traceback import TracebackException

LINES = '-' * 80
NOT_FOUND = '{} Not found'
CREATED = '{} Created'
EXISTS = '{} Exists'
DELETED = '{} Deleted'
MOVED = '{} Moved to {}'
COPIED = '{} Copied to {}'
SAVED = '{} Saved'
LOADED = '{} Loaded'
UPLOADED = '{} Uploaded to {}'
DOWNLOADED = '{} Downloaded to {}'
CONTENT = 'Content: {}'
SUCCESS_C = 'light_green'
SUCCESS_C2 = ['black', 'bold', 'background_green']
FAIL_C = 'red'
FAIL_C2 = ['black', 'bold', 'background_red']

CONST_COLOR_SHORTCUTS = {
    'b': 'blue',
    'g': 'green',
    'r': 'red',
    'c': 'cyan',
    'm': 'magenta',
    'y': 'yellow',
    'k': 'black',
    'w': 'white',
    'bo': 'bold',
    'un': 'underlined',
    're': 'reverse',
}

CONST_COLOR_MAP = {
    'reset_all': "\033[0m",
    'bold': "\033[1m",
    'underlined': "\033[4m",
    'reverse': "\033[7m",  # switch font color and background color

    'blue': "\033[34m",
    'background_blue': "\033[44m",
    'light_blue': "\033[94m",
    'background_light_blue': "\033[104m",

    'green': "\033[32m",
    'background_green': "\033[42m",
    'light_green': "\033[92m",
    'background_light_green': "\033[102m",

    'red': "\033[31m",
    'background_red': "\033[41m",
    'light_red': "\033[91m",
    'background_light_red': "\033[101m",

    'cyan': "\033[36m",
    'background_cyan': "\033[46m",
    'background_light_cyan': "\033[106m",
    'light_cyan': "\033[96m",

    'magenta': "\033[35m",
    'background_magenta': "\033[45m",
    'light_magenta': "\033[95m",
    'background_light_magenta': "\033[105m",

    'yellow': "\033[33m",
    'background_yellow': "\033[43m",
    'light_yellow': "\033[93m",
    'background_light_yellow': "\033[103m",

    'light_gray': "\033[37m",
    'background_light_gray': "\033[47m",
    'dark_gray': "\033[90m",
    'background_dark_gray': "\033[100m",

    'black': "\033[97m",
    'background_black': "\033[107m",

    'white': "\033[30m",
    'background_white': "\033[40m",
}


def get_linkable_exception() -> str:
    """
    :return: e.g. File "D:\workspace\2021wizzi_utils\wizzi_utils\misc\test\test_misc_tools.py", line 851,
                            in get_linkable_exception_test: division by zero
    where file is clickable
    see get_linkable_exception_test
    """
    _, value, tb = sys.exc_info()
    info_str = ''
    for i, line in enumerate(TracebackException(type(value), value, tb, limit=None).format(chain=True)):
        if "File \"" in line:  # care only for the line with: filename(clickable), line, func name
            # print(i, line.split('\n')[0])
            info_str = line.split('\n')[0].strip()  # remove the exception itself
    return info_str


def exception_error(e: (Exception, str), real_exception: bool = False, depth: int = 2, tabs: int = 1):
    """
    Aux function - print exception error in red with function name
    :param e: error. e.g. <class 'ModuleNotFoundError'> or str of your choosing
    :param real_exception: if True: gets info from sys. else, manually get file and line
    :param depth: if real_exception False, needs the depth to get file and line
    :param tabs:
    :return:
    see get_linkable_exception_test
    """
    if real_exception:
        err_meta = get_linkable_exception()
    else:
        err_meta = get_function_name_and_line(depth=depth)
    error_str = '{}{}: {}'.format(tabs * '\t', err_meta, e)
    print(add_color(string=error_str, ops='Red'))
    return


def chop_microseconds(delta: datetime.timedelta) -> datetime.timedelta:
    """
    Aux function - removes micro seconds from datetime.timedelta object
    e.g. 0:00:02.000001 -> 0:00:02
    :param delta:
    :return:
    """
    return delta - datetime.timedelta(microseconds=delta.microseconds)


def get_timer_delta(s_timer: float, e_timer: float = None, with_ms: bool = False, ack: bool = False,
                    tabs: int = 1) -> datetime.timedelta:
    """
    :param s_timer: begin time
    :param e_timer: end time - optional. if None end time is now
    :param with_ms: if microseconds needed - set to true. else: no microseconds
    :param ack: print time passed
    :param tabs:
    :return:
    see timer_test()
    """
    if e_timer is None:
        e_timer = get_timer()
    d = datetime.timedelta(seconds=(e_timer - s_timer))
    if not with_ms:
        d = chop_microseconds(d)
    if ack:
        print('{}Time passed {}'.format(tabs * '\t', d))
    return d


def get_timer() -> float:
    """
    sets a timer beginning
    :return:
    see timer_test()
    """
    return timer()


def timer_action(seconds: int, action: str = '', tabs: int = 1) -> None:
    """
    :param seconds:
    :param action:
    :param tabs:
    :return:
    counts till seconds or block
    see timer_action_test()
    """
    if seconds is None:
        input('{}Press "Enter" key for {}...'.format(tabs * '\t', action))
    else:
        time_in_future = datetime.datetime.now() + datetime.timedelta(seconds=seconds)
        print('{}{} IN: {}'.format(tabs * '\t', action, seconds), end='', flush=True)
        while time_in_future > datetime.datetime.now():
            time.sleep(1)
            seconds -= 1
            print(' {}'.format(seconds), end='', flush=True)
        print('')
    return


def get_time_stamp(format_s: str = '%Y_%m_%d_%H_%M_%S', ack: bool = False, tabs: int = 1) -> str:
    """
    :param format_s: date time format
    :param ack: prints current time
    :param tabs:
    the default is for files
    :return:
    see get_current_date_hour_test()
    """
    now = datetime.datetime.now()
    time_stamp = now.strftime(format_s)
    if ack:
        print('{}timeStamp = {}'.format(tabs * '\t', time_stamp))
    return time_stamp


def get_pc_name(ack: bool = False, tabs: int = 1) -> str:
    """
    :param ack:
    :param tabs:
    :return: pc name as str
    see get_pc_name_test()
    """
    try:
        pc_name = platform.uname()[1]
        if ack:
            print('{}* Computer Name: {}'.format(tabs * '\t', pc_name))
    except ModuleNotFoundError as e:
        pc_name = ''
        exception_error(e, real_exception=True)
    return pc_name


def get_mac_address(with_colons: bool = True, ack: bool = False, tabs: int = 1) -> str:
    """
    :return: pc mac as str
    see get_mac_address_test()
    """
    try:
        mac = get_mac()
        mac = "{:X}".format(mac)  # turn to Hex
        if with_colons:
            mac = ':'.join(mac[i:i + 2] for i in range(0, 12, 2))
        if ack:
            print('{}* Computer Mac: {}'.format(tabs * '\t', mac))
    except ModuleNotFoundError as e:
        mac = ''
        exception_error(e, real_exception=True)
    return mac


def get_cuda_version(ack: bool = False, tabs: int = 1) -> str:
    """
    :param ack:
    :param tabs:
    :return: cuda version if found on environment variables
    see get_cuda_version_test()
    """
    cuda_path = get_env_variable(key='CUDA_PATH')
    if cuda_path is not None:
        cuda_v = add_color('{}* CUDA Version: {}'.format(tabs * '\t', os.path.basename(cuda_path)), ops=SUCCESS_C)
    else:
        cuda_v = add_color('* No CUDA_PATH found', ops=FAIL_C)
    if ack:
        print(cuda_v)
    return cuda_v


def get_env_variables(ack: bool = False, tabs: int = 1) -> dict:
    """
    :param ack:
    :param tabs:
    :return: dict with envs
    see get_env_variables_test()
    """
    env_d = dict(os.environ)
    if ack:
        print('{}Environment variables:'.format(tabs * '\t'))
        for k, v in env_d.items():
            print('{}\t{} = {}'.format(tabs * '\t', k, v))
    return env_d


def set_env_variable(key: str, val: str, ack: bool = False, tabs: int = 1) -> None:
    """
    :param key:
    :param val:
    :param ack:
    :param tabs:
    insert new env variable
    see set_env_variable_test()
    """
    key = key.upper()
    os.environ[key] = val
    if ack:
        print('{}Inserted to environment: {} = {}'.format(tabs * '\t', key, val))
    return


def get_env_variable(key: str, ack: bool = False, tabs: int = 1) -> str:
    """
    :param key:
    :param ack:
    :param tabs:
    :return: env variable value
    see get_env_variable_test()
    """
    key = key.upper()
    ret_val = os.environ[key] if key in os.environ else None
    if ack:
        if ret_val is not None:
            print('{}{} = {}'.format(tabs * '\t', key, ret_val))
        else:
            exception_error(NOT_FOUND.format(key), real_exception=False, tabs=tabs)
    return ret_val


def del_env_variable(key: str, ack: bool = False, tabs: int = 1) -> None:
    """
    :param key:
    :param ack:
    :param tabs:
    :return: env variable value
    see get_env_variable_test()
    """
    key = key.upper()
    if key in os.environ:
        del os.environ[key]
        if ack:
            print('{}{}'.format(tabs * '\t', DELETED.format(key)))
    else:
        exception_error(NOT_FOUND.format(key), real_exception=False, tabs=tabs)
    return


def silence_cv_warn() -> None:
    """
    disables on windows:
    [ WARN:0] global .... src\cap_msmf.cpp (438)
    `anonymous-namespace'::SourceReaderCB::~SourceReaderCB terminating async callback
    :return:
    if doesn't work on linux try:
    set_env_variable(key='OPENCV_VIDEOIO_DEBUG', val='0')
    """
    set_env_variable(key='OPENCV_VIDEOIO_PRIORITY_MSMF', val='0')
    return


def make_cuda_invisible() -> None:
    """
        disable gpu 0
        FUTURE - support disabling many GPUS
        os.environ['CUDA_VISIBLE_DEVICES'] = '0'  # GPU 0 available
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1, 0'  # GPU 0 not available
        os.environ['CUDA_VISIBLE_DEVICES'] = '0, -1, 0'  # GPU 0 available but 1 disabled
        os.environ['CUDA_VISIBLE_DEVICES'] = '1, 2, -1, 0'  # GPU 1,2 available but 0 disabled

        read more: https://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#env-vars
        see make_cuda_invisible_test()
    """
    set_env_variable(key='CUDA_VISIBLE_DEVICES', val='-1, 0')
    return


def start_profiler() -> cProfile.Profile:
    """
    starts profiling
    :return: profiling object that is needed for end_profiler()
    see profiler_test()
    """
    pr = cProfile.Profile()
    pr.enable()
    return pr


def end_profiler(pr: cProfile.Profile, rows: int = 10, ack: bool = False) -> str:
    """
    profiling output
    :param pr: object returned from start_profiler()
    :param rows: how many rows to print sorted by 'cumulative' run time
    :param ack:
    :return: profiler output as string
    see profiler_test()
    """
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
    ps.print_stats(rows)
    profiler_str = s.getvalue()
    if ack:
        print('{}'.format(profiler_str))
    return profiler_str


def set_seed(seed: int = 42) -> None:
    """
    :param seed: setting numpy and random seeds
    :return:
    see main_wrapper_test() - uses set_seed
    """
    np.random.seed(seed)
    random.seed(seed)
    return


def version() -> str:
    """
    :return:
    """
    # v = pkg_resources.require("wizzi_utils")[0].version
    v = '6.6.6'
    return v


def is_macos():
    """
    :return:
    see os_test()
    TODO *** NOT TESTED
    """
    return platform.system() == "Darwin"


def is_windows():
    """
    :return:
    see os_test() - tested
    """
    return platform.system() == "Windows"


def is_linux():
    """
    :return:
    see os_test()
    """
    return platform.system() == "Linux"


def is_raspi_arm():
    """
    armv7l is 32 bit processor.
    :return:
    see os_test()
    """
    try:
        res = os.uname()[4] == "armv7l"
    except AttributeError:
        res = False
    return res


def is_aarch64():
    """
    ARM64 is the 64-bit extension of the ARM architecture
    :return:
    see os_test()
    TODO *** NOT TESTED
    TODO if doesn't work - https://en.wikipedia.org/wiki/Uname
    """
    try:
        res = os.uname()[4] == "aarch64"
    except AttributeError:
        res = False
    return res


def is_nano():
    """
    :return:
    TODO *** NOT TESTED
    see os_test()
    """
    try:
        res = os.uname()[1] == "nano"
    except AttributeError:
        res = False
    return res


def get_system_info() -> str:
    """
    print(sys.platform):
    https://docs.python.org/3/library/sys.html#sys.platform
        AIX 'aix'
        Linux 'linux'
        Windows 'win32'
        Windows/Cygwin 'cygwin'
        macOS 'darwin'

    print(os.name):
        Windows 'nt'
        RPi 'posix' # maybe for all linux
    print(platform.system()):
        Windows 'Windows'
    :return:
    """
    named_tuple = platform.uname()
    # t = tuple(named_tuple)
    return str(named_tuple)


def get_wizzi_utils_root() -> str:
    """
    :return:
    """
    # project_root/misc/misc_tools.py -> remove the py and the misc folder
    project_root = os.path.dirname(os.path.dirname(get_file_name()))
    return project_root


def main_wrapper(
        main_function: Callable,
        seed: int = -1,
        ipv4: bool = False,
        cuda_off: bool = False,
        torch_v: bool = False,
        tf_v: bool = False,
        cv2_v: bool = False,
        with_profiler: bool = False
) -> None:
    """
    :param main_function: the function to run
    :param seed: if -1 no seed, else set_seed(seed=seed)
    :param ipv4: print computer ipv4
    :param cuda_off: make gpu invisible and force run on cpu
    :param torch_v: print torch version
    :param tf_v: print tensorflow version
    :param cv2_v: print opencv version
    :param with_profiler: run profiler
    template:
    wu.main_wrapper(
        main_function=main,
        seed=42,
        ipv4=False,
        cuda_off=False,
        torch_v=False,
        tf_v=False,
        cv2_v=False,
        with_profiler=False
    )
    see main_wrapper_test()
    :return:
    """
    print(LINES)
    start_timer = get_timer()

    print('main_wrapper:')
    print('* Run started at {}'.format(get_time_stamp(format_s='%d-%m-%Y %H:%M:%S')))
    print('* Python Version {}'.format(sys.version))
    print('* Operating System {}'.format(get_system_info()))
    print('* Interpreter: {}'.format(sys.executable))
    print('* wizzi_utils Version {}'.format(version()))
    print('* Working Dir: {}'.format(os.getcwd()))
    print('* Computer Mac: {}'.format(get_mac_address()))
    print('* CPU Info: {}'.format(cpu_info(one_liner=True, tabs=0)))
    print('* Physical Memory: {}'.format(hard_disc(one_liner=True, tabs=0)))
    print('* RAM: {}'.format(ram_size(one_liner=True, tabs=0)))

    if ipv4:
        try:
            from wizzi_utils.socket.socket_tools import get_ipv4
            print('* Computer ipv4: {}'.format(get_ipv4()))
        except (ImportError, ModuleNotFoundError, NameError) as err:
            print(add_color('* {}'.format(err), 'r'))

    cuda_msg = get_cuda_version(ack=False, tabs=0)
    if cuda_off:
        make_cuda_invisible()
        cuda_msg += ' (Turned off)'
    print(cuda_msg)

    if torch_v:
        try:
            from wizzi_utils.torch.torch_tools import get_torch_version
            print(get_torch_version(ack=False, tabs=0))
        except (ImportError, ModuleNotFoundError, NameError) as err:
            print(add_color('* {}'.format(err), 'r'))

    if cv2_v:
        try:
            from wizzi_utils.open_cv.open_cv_tools import get_cv_version
            print(get_cv_version(ack=False, tabs=0))
        except (ImportError, ModuleNotFoundError, NameError) as err:
            pip = 'pip install opencv-python'
            print(add_color('* {} - {}'.format(err, pip), 'r'))

    if tf_v:
        try:
            from wizzi_utils.tensorflow.tensorflow_tools import get_tensorflow_version
            print(get_tensorflow_version(ack=False, tabs=0))
        except (ImportError, ModuleNotFoundError, NameError) as err:
            print(add_color('* {}'.format(err), 'r'))

        try:
            from wizzi_utils.tflite.tflite_tools import get_tflite_version
            print(get_tflite_version(ack=False, tabs=0))
        except (ImportError, ModuleNotFoundError, NameError) as err:
            pip = '\n\twin os: pip install --extra-index-url https://google-coral.github.io/py-repo/ tflite_runtime'
            print(add_color('* {} - {}'.format(err, pip), 'r'))

    if seed > -1:
        set_seed(seed=seed)
        print('* Seed was initialized to {}'.format(seed))

    print('Function {} started:'.format(main_function))
    print(LINES)
    pr = start_profiler() if with_profiler else None

    main_function()
    if with_profiler:
        print(end_profiler(pr))

    print(LINES)
    print('Total run time {}'.format(get_timer_delta(start_timer)))
    return


def get_data_str(data_str_raw, chars: (int, str)) -> str:
    """
    Aux function for to_str()
    :param data_str_raw:
    :param chars:
    :return: data_str
    """
    data_str_raw = data_str_raw.replace('\n', '').replace('  ', '')
    if chars == 'all':  # all data
        chars = len(data_str_raw) + 1
    elif chars is None:  # no data
        chars = 0

    data_str_rep = ': {}'.format(data_str_raw[:chars])

    if len(data_str_raw) > chars > 0:
        data_str_rep += ' ...too long'
    return data_str_rep


def is_int(var: any) -> bool:
    return isinstance(var, (int, np.int, np.int32, np.uint8))


def is_float(var: any) -> bool:
    return isinstance(var, (float, np.float, np.float32, np.float64))


def is_str(var: any) -> bool:
    return isinstance(var, str)


def is_list(var: any) -> bool:
    return isinstance(var, list)


def is_tuple(var: any) -> bool:
    return isinstance(var, tuple)


def is_numpy(var: any) -> bool:
    return isinstance(var, np.ndarray)


def is_dict(var: any) -> bool:
    return isinstance(var, dict)


def to_str(var: any,
           title: str = 'var',
           chars: (int, str) = 100,
           fp: (int, None) = 2,
           wm: bool = True,
           rec: bool = False
           ) -> str:
    """
    :param var: the variable
    :param title: str: the title (usually variable name)
    :param chars: int, None or str:
        chars>0: maximal number of chars
        chars==None: no chars
        chars=='all': all chars
    :param fp: float_precision: round number if possible(float, list or np array of floats...)
            fp>=0 round
            fp==None: no rounding
    :param wm: with_meta: with meta data such as type, len\shape, dtype...
    :param rec: recursive: to keep printing if there are more items inside e.g. np.array(shape=(2,3,4)) -> 3 prints
    :return: informative string of the variable
    see to_str_test()
    """

    string = title

    if is_float(var):
        if wm:
            type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name
            string += '({})'.format(type_s)
        if fp is not None and fp >= 0:
            f_format = '{:,.%xf}' % fp
        else:
            f_format = '{:,}'
        data_str_raw = f_format.format(var)
        string += get_data_str(data_str_raw, chars)

    elif is_int(var):
        if wm:
            type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name
            string += '({})'.format(type_s)
        data_str_raw = '{:,}'.format(var)
        string += get_data_str(data_str_raw, chars)

    elif is_str(var):
        if wm:
            type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name
            string += '({}'.format(type_s)
            string += ',len={})'.format(var.__len__())
        string += get_data_str(var, chars)

    elif is_list(var) or is_tuple(var):
        if wm:
            type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name
            string += '({}'.format(type_s)
            string += ',len={})'.format(var.__len__())

        if fp is not None and fp >= 0:
            # 1d list\tuple - nice print
            if all((is_int(item) or is_float(item)) for item in var):
                if fp > 0 and all(is_int(item) for item in var):
                    fp = 0
                f_format = '{:,.%xf}' % fp
                new_v = [f_format.format(li_item) for li_item in var]
            else:  # 2d list\tuple - could fail if found strings or types that cant be rounded
                new_v = round_list(var, fp)
        else:  # fp = None - no rounding
            new_v = var

        if is_tuple(var):
            new_v = tuple(new_v)

        data_str_raw = str(new_v).replace('\'', '')

        string += get_data_str(data_str_raw, chars)
        if rec and len(var) > 0:
            inner_str = to_str(var=var[0], title='{}[0]'.format(title), chars=chars, fp=fp, rec=rec)
            string += '\n\t{}'.format(inner_str)

    elif is_numpy(var):
        if wm:
            type_s = str(type(var)).replace('<class \'', '').replace('\'>', '').replace('numpy.', '')
            string += '({}'.format(type_s)
            string += ',s={}'.format(var.shape)
            string += ',dtype={})'.format(var.dtype)

        if fp is not None and fp >= 0:
            new_v = np.round(var.tolist(), fp).tolist()
        else:
            new_v = var.tolist()
        data_str_raw = str(new_v).replace('\'', '')
        string += get_data_str(data_str_raw, chars)
        if rec and var.shape[0] > 0:  # recursive call
            inner_str = to_str(var=var[0], title='{}[0]'.format(title), chars=chars, fp=fp, rec=rec)
            string += '\n\t{}'.format(inner_str)

    elif is_dict(var):
        if wm:
            type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name
            string += '({}'.format(type_s)
            string += ',len={}'.format(var.__len__())
            string += ',keys={})'.format(list(var.keys()))
        string += get_data_str(str(var), chars)
        if rec and len(var) > 0:  # recursive call
            first_k = next(iter(var))
            inner_str = to_str(var=var[first_k], title='{}[{}]'.format(title, first_k), chars=chars, fp=fp, rec=rec)
            string += '\n\t{}'.format(inner_str)

    else:
        # all unidentified elements get default print (title(type): data)
        # must have str() to this type
        if wm:
            type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name
            string += '({})'.format(type_s)
        string += get_data_str(str(var), chars)
    return string


def save_np(t: np.array, path: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param t: numpy array
    :param path: suffix '.npy' added automatically if not exists
    :param ack:
    :param tabs:
    :return:
    see save_load_np_test()
    """
    np.save(path, t)
    if ack:
        size_s = file_or_folder_size(path)
        file_msg = '{}({})'.format(path, size_s)
        print('{}{}'.format(tabs * '\t', SAVED.format(file_msg)))
    return


def load_np(path: str, ack: bool = True, tabs: int = 1) -> np.array:
    """
    :param path:
    :param ack:
    :param tabs:
    :return: numpy array
    see save_load_np_test()
    """
    if os.path.exists(path):
        t = np.load(path)
        if ack:
            size_s = file_or_folder_size(path)
            file_msg = '{}({})'.format(path, size_s)
            print('{}{}'.format(tabs * '\t', LOADED.format(file_msg)))
    else:
        exception_error(NOT_FOUND.format(path), real_exception=False, tabs=tabs)
        t = None
    return t


def save_npz(arrays_dict: dict, path: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param arrays_dict: e.g. { 'a': np.ones(3) }
    :param path:
    :param ack:
    :param tabs:
    :return:
    save a dict of numpy arrays
    see save_load_npz_test()
    """
    np.savez(path, **{n: a for n, a in arrays_dict.items()})
    if ack:
        size_s = file_or_folder_size(path)
        file_msg = '{}({})'.format(path, size_s)
        print('{}{}. Keys={}'.format(tabs * '\t', SAVED.format(file_msg), arrays_dict.keys()))
    return


def load_npz(path: str, ack: bool = True, tabs: int = 1) -> dict:
    """
    :param path:
    :param ack:
    :param tabs:
    :return: numpy array
    see save_load_npz_test()
    """
    if os.path.exists(path):
        arrays_obj = np.load(path)
        arrays_dict = {}
        # noinspection PyUnresolvedReferences
        for k in arrays_obj.files:
            arrays_dict[k] = arrays_obj[k]
        if ack:
            size_s = file_or_folder_size(path)
            file_msg = '{}({})'.format(path, size_s)
            print('{}{}. Keys={}'.format(tabs * '\t', LOADED.format(file_msg), arrays_dict.keys()))
    else:
        exception_error(NOT_FOUND.format(path), real_exception=False, tabs=tabs)
        arrays_dict = None
    return arrays_dict


def save_pkl(data_dict: dict, path: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param data_dict:
    :param path:
    :param ack:
    :param tabs:
    :return:
    see save_load_pkl_test()
    """
    file_obj = open(path, "wb")
    pickle.dump(data_dict, file_obj)
    file_obj.close()
    if ack:
        size_s = file_or_folder_size(path)
        file_msg = '{}({})'.format(path, size_s)
        print('{}{}'.format(tabs * '\t', SAVED.format(file_msg)))
    return


def load_pkl(path: str, ack: bool = True, tabs: int = 1) -> dict:
    """
    :param path:
    :param ack:
    :param tabs:
    :return:
    see save_load_pkl_test()
    """
    if os.path.exists(path):
        file_obj = open(path, "rb")
        data_dict = pickle.load(file_obj)
        file_obj.close()
        if ack:
            size_s = file_or_folder_size(path)
            file_msg = '{}({})'.format(path, size_s)
            print('{}{}'.format(tabs * '\t', LOADED.format(file_msg)))
    else:
        exception_error(NOT_FOUND.format(path), real_exception=False, tabs=tabs)
        data_dict = None
    return data_dict


def get_uniform_dist_by_dim(A: [np.array, list]) -> (np.array, np.array):
    """
    :param A:
    :return:
    for every dimension gets the lowest and highest
    see get_uniform_dist_by_dim_test()
    """
    lows = np.min(A, axis=0)
    highs = np.max(A, axis=0)
    return lows, highs


def get_normal_dist_by_dim(A: [np.array, list]) -> (np.array, np.array):
    """
    :param A:
    :return:
    see get_normal_dist_by_dim_test()
    """
    means = np.mean(A, axis=0)
    stds = np.std(A, axis=0)
    return means, stds


def np_uniform(shape: tuple, lows: [list, int], highs: [list, int]) -> np.array:
    """
    :param shape:
    :param lows:
    :param highs:
    :return:
    see np_uniform_test()
    """
    ret = np.random.uniform(low=lows, high=highs, size=shape)
    return ret


def np_normal(shape: tuple, mius: [list, int, float], stds: [list, int, float]) -> np.array:
    """
    :param shape:
    :param mius:
    :param stds:
    :return:
    see np_normal_test()
    """
    ret = np.random.normal(loc=mius, scale=stds, size=shape)
    return ret


def generate_new_data_from_old(old_data: np.array, new_data_n: int, dist: str = 'normal'):
    """
    :param old_data:
    :param new_data_n:
    :param dist:
    :return:
    see generate_new_data_from_old_test()
    """
    d = old_data.shape[1]
    if dist == 'uniform':
        lows, highs = get_uniform_dist_by_dim(old_data)
        new_data = np_uniform(shape=(new_data_n, d), lows=lows, highs=highs)
    else:  # else normal
        means, stds = get_normal_dist_by_dim(old_data)
        new_data = np_normal(shape=(new_data_n, d), mius=means, stds=stds)
    return new_data


def np_random_integers(low: int, high: int, size: tuple) -> np.array:
    """
    :param low:
    :param high:
    :param size:
    :return:
    see np_random_integers_test()
    """
    ret = np.random.random_integers(low=low, high=high, size=size)
    return ret


def augment_x_y_numpy(X: np.array, y: np.array) -> np.array:
    """
    :param X:
    :param y:
    :return:
    see augment_x_y_numpy_test()
    """
    assert X.shape[0] == y.shape[0], 'row count must be the same'
    if len(X.shape) == 1:  # change x size()=[n,] to size()=[n,1]
        X = X.reshape(X.shape[0], 1)
    if len(y.shape) == 1:  # change y size()=[n,] to size()=[n,1]
        y = y.reshape(y.shape[0], 1)
    A = np.column_stack((X, y))
    return A


def de_augment_numpy(A: np.array) -> (np.array, np.array):
    """
    :param A:
    :return:
    see de_augment_numpy_test()
    """
    if len(A.shape) == 1:  # A is 1 point. change from size (n) to size (1,n)
        A = A.reshape(1, A.shape[0])
    X, y = A[:, :-1], A[:, -1]
    if len(X.shape) == 1:  # change x size()=[n,] to size()=[n,1]
        X = X.reshape(X.shape[0], 1)
    if len(y.shape) == 1:  # change y size()=[n,] to size()=[n,1]
        y = y.reshape(y.shape[0], 1)
    return X, y


def nCk(n: int, k: int, as_int: bool = False):
    """
    :param n:
    :param k:
    :param as_int:
    :return: if as_int True: the result of nCk, else the combinations of nCk
    n choose k
    see nCk_test()
    """
    range_list = np.arange(0, n, 1)
    combs = list(combinations(range_list, k))
    combs = [list(comb) for comb in combs]
    if as_int:
        combs = len(combs)
    return combs


def redirect_std_start() -> (io.TextIOWrapper, io.StringIO):
    """
    redirect all prints to summary_str
    :return:
        io.TextIOWrapper - to revert back the prints to sys.stdout
        io.StringIO - to extract output
    see redirect_std_test()
    """
    old_stdout = sys.stdout
    sys.stdout = summary_str = io.StringIO()
    return old_stdout, summary_str


def redirect_std_finish(old_stdout: io.TextIOWrapper, summary_str: io.StringIO) -> str:
    """
    :param old_stdout: to revert back the prints to sys.stdout
    :param summary_str: to extract output
    :return:
    redirect all prints back to std out and return a string of what was captured"
    see redirect_std_test()
    """
    sys.stdout = old_stdout
    return summary_str.getvalue()


def get_line_number(depth: int = 1, ack: bool = False, tabs: int = 1) -> str:
    """
    :param depth:
    :param ack:
    :param tabs:
    :return:
    see get_line_number_test()
    """
    ret_val = ''
    try:
        scope_1_back = inspect.stack()[depth]  # stack()[0] is this function
        ret_val = '{}'.format(scope_1_back.lineno)
        if ack:
            print('{}Line {}:'.format(tabs * '\t', ret_val))
    except IndexError as e:
        exception_error(e, real_exception=True)
    return ret_val


def get_function_name(depth: int = 1, ack: bool = False, tabs: int = 1) -> str:
    """
    :param depth:
    :param ack:
    :param tabs:
    :return:
    see
    """
    ret_val = ''
    try:
        scope_1_back = inspect.stack()[depth]  # stack()[0] is this function
        ret_val = '{}'.format(scope_1_back.function)
        if ack:
            print('{}{}:'.format(tabs * '\t', ret_val))
    except IndexError as e:
        exception_error(e, real_exception=True)
    return ret_val


def get_file_name(depth: int = 1, ack: bool = False, tabs: int = 1) -> str:
    """
    :param depth:
    :param ack:
    :param tabs:
    :return:
    see get_file_name_test()
    """
    ret_val = ''
    try:
        scope_1_back = inspect.stack()[depth]  # stack()[0] is this function
        ret_val = '{}'.format(scope_1_back.filename)
        if ack:
            print('{}{}:'.format(tabs * '\t', ret_val))
    except IndexError as e:
        exception_error(e, real_exception=True)
    return ret_val


def get_base_file_name(depth: int = 1, ack: bool = False, tabs: int = 1) -> str:
    """
    :param depth:
    :param ack:
    :param tabs:
    :return:
    see get_base_file_name_test()
    """
    # +1 because of this function
    file_name = get_file_name(depth + 1)
    base_name = os.path.basename(file_name)
    if ack:
        print('{}{}:'.format(tabs * '\t', base_name))
    return base_name


def get_function_name_and_line(depth: int = 1, ack: bool = False, tabs: int = 1) -> str:
    """
    :param depth:
    :param ack:
    :param tabs:
    :return:
    see get_function_name_and_line_test()
    """
    # +1 because of this function
    ret_val = '{}::{}'.format(get_function_name(depth + 1), get_line_number(depth + 1))
    if ack:
        print('{}{}:'.format(tabs * '\t', ret_val))
    return ret_val


def get_base_file_and_function_name(depth: int = 1, ack: bool = False, tabs: int = 1) -> str:
    """
    :param depth:
    :param ack:
    :param tabs:
    :return:
    see get_base_file_and_function_name_test()
    """
    # +1 because of this function
    ret_val = '{}::{}'.format(get_base_file_name(depth + 1), get_line_number(depth + 1))
    if ack:
        print('{}{}:'.format(tabs * '\t', ret_val))
    return ret_val


def add_color(
        string: str,
        ops: (str, list) = 'red'
) -> str:
    """
    :param string:
    :param ops:
        str: color, bg_color, bold, underline, reverse
        list: every combination of color, bg_color, bold, underline, reverse
    colors:
    * X = { blue(b), green(g), red(r), cyan(c), magenta(m), yellow(y) }
        for x in X: x, background_x, light_x, background_light_x
    * more colors:
        light_gray, background_light_gray, dark_gray, background_dark_gray
        black, background_black
        white, background_white

    special options:
    * bold(bo), underlined(un), reverse(re)

    to see all colors and options:
        for k, v in wu.CONST_COLOR_MAP.items():
            print('{}{}{}'.format(v, k, wu.CONST_COLOR_MAP['reset_all']))
    see add_color_test()
    """
    out_string = string
    # first to lower and check shortcuts
    if is_list(ops):
        pass
    elif is_str(ops):
        ops = [ops]
    colors = ''
    for op in ops:
        op = op.lower()
        if op in CONST_COLOR_SHORTCUTS:
            op = CONST_COLOR_SHORTCUTS[op]
        if op in CONST_COLOR_MAP:
            colors += CONST_COLOR_MAP[op]
    if colors != '':
        out_string = '{}{}{}'.format(colors, string, CONST_COLOR_MAP['reset_all'])

    return out_string


# noinspection PyTypeChecker
def init_logger(logger_path: str = './log.txt') -> io.TextIOWrapper:
    """
    :param: logger_path
    :return: logger obj
    see logger_test()
    """
    global logger
    logger = open(file=logger_path, mode='w', encoding='utf-8')
    return logger


def flush_logger() -> None:
    """
    good for loops - writes every iteration if used
    see logger_test()
    """
    global logger
    if logger is not None:
        logger.flush()
    return


def log_print(line: str, tabs: int = 0) -> None:
    """
    :param line:
    :param tabs:
    :return:
    see logger_test()
    """
    global logger
    print('{}{}'.format(tabs * '\t', line))
    if logger is not None:
        logger.write('{}{}\n'.format(tabs * '\t', line))
    return


def log_print_dict(my_dict, tabs: int = 1) -> None:
    """
    :param my_dict:
    :param tabs:
    :return:
    see logger_test()
    """
    for k, v in my_dict.items():
        log_print('{}{}: {}'.format(tabs * '\t', k, v))
    return


def close_logger() -> None:
    """
    :return:
    see logger_test()
    """
    global logger
    if logger is not None:
        logger.close()
    return


def create_dir(dir_path: str, ack: bool = True, tabs: int = 1):
    """
    :param dir_path:
    :param ack:
    :param tabs:
    :return:
    see create_and_delete_dir_test()
    """
    if not os.path.exists(dir_path):
        try:
            # os.mkdir(dir_path)
            Path(dir_path).mkdir(parents=True, exist_ok=True)
            if ack:
                print('{}{}'.format(tabs * '\t', CREATED.format(dir_path)))
        except OSError as e:
            exception_error(e, real_exception=True)
    else:
        exception_error(EXISTS.format(dir_path), real_exception=False, tabs=tabs)
    return


def delete_empty_dir(dir_path: str, ack: bool = True, tabs: int = 1):
    """
    :param dir_path:
    :param ack:
    :param tabs:
    :return:
    see create_and_delete_empty_dir_test()
    """
    if os.path.exists(dir_path):
        files_and_dirs = []
        for _dir_path, dir_names, file_names in os.walk(dir_path):
            for d in dir_names:
                # fd = '{}/{}'.format(_dir_path.replace('\\', '/'), d)
                files_and_dirs.append(d)
            for f in file_names:
                # fp = '{}/{}'.format(_dir_path.replace('\\', '/'), f)
                files_and_dirs.append(f)
        if len(files_and_dirs) > 0:
            err = '{} HAS {} FILES\DIRS {} - use delete_dir_with_files()'.format(dir_path, len(files_and_dirs),
                                                                                 files_and_dirs)
            exception_error(err, real_exception=False, tabs=tabs)
        else:
            try:
                os.rmdir(dir_path)
                if ack:
                    print('{}{}'.format(tabs * '\t', DELETED.format(dir_path)))
            except OSError as e:
                exception_error(e, real_exception=True)
    else:
        exception_error(NOT_FOUND.format(dir_path), real_exception=False, tabs=tabs)
    return


def delete_dir_with_files(dir_path: str, ack: bool = True, tabs: int = 1):
    """
    :param dir_path:
    :param ack:
    :param tabs:
    :return:
    see create_and_delete_dir_test()
    """
    if os.path.exists(dir_path):
        files_and_dirs = []
        for _dir_path, dir_names, file_names in os.walk(dir_path):
            for d in dir_names:
                # fd = '{}/{}'.format(_dir_path.replace('\\', '/'), d)
                files_and_dirs.append(d)
            for f in file_names:
                # fp = '{}/{}'.format(_dir_path.replace('\\', '/'), f)
                files_and_dirs.append(f)
        try:
            size_s = file_or_folder_size(dir_path)
            shutil.rmtree(dir_path)
            status = '({} files\dirs, size {} - {})'.format(len(files_and_dirs), size_s, files_and_dirs)
            if ack:
                print('{}{} - {}'.format(tabs * '\t', DELETED.format(dir_path), status))
        except OSError as e:
            exception_error(e, real_exception=True)
    else:
        exception_error(NOT_FOUND.format(dir_path), real_exception=False, tabs=tabs)
    return


def find_files_in_folder(dir_path: str, file_suffix: str, ack: bool = False, tabs: int = 1) -> list:
    """
    :param dir_path:  root dir. will be looked in recursively
    :param file_suffix:  e.g. '.jgp'
    :param ack:
    :param tabs:
    :return:
    see find_files_in_folder_test()

    sort by numbers in basename:
    folder_path = './path' # content [1.jpg, 10.jpg 2.jpg]
    folder_imgs=wu.find_files_in_folder(folder_path, file_suffix='.jpg', ack=True)
    print(folder_imgs)
    # >>> [1.jpg, 10.jpg, 2.jpg]
    def sorter(x):
        return int(os.path.basename(x).split('.')[0])
    folder_imgs.sort(key=sorter)
    print(folder_imgs)
    # >>> [1.jpg, 2.jpg, 10.jpg]

    folder_path = './path' # content [gl_1.jpg, gl_10.jpg gl_2.jpg]
    folder_imgs=wu.find_files_in_folder(folder_path, file_suffix='.jpg', ack=True)
    print(folder_imgs)
    # >>> [gl_1.jpg, gl_10.jpg, gl_2.jpg]
    def sorter(x):
        return int(os.path.basename(x).split('.')[0][3:])
    folder_imgs.sort(key=sorter)
    print(folder_imgs)
    # >>> [gl_1.jpg, gl_2.jpg, gl_10.jpg]
    """

    if os.path.exists(dir_path):
        dir_path = os.path.abspath(dir_path)  # to get output of full absolute path
        all_files_found = glob.glob('{}/**/*{}'.format(dir_path, file_suffix), recursive=True)

        all_files_processed = []
        for fp in all_files_found:
            fp = fp.replace('\\', '/')  # replace windows folder \ with /
            if fp.startswith('./'):  # e.g. ./test.txt -> test.txt
                fp = fp[2:]
            all_files_processed.append(fp)
        if ack:
            all_files_base = [os.path.basename(fp) for fp in all_files_processed]
            print('{}found {} files that ends with {} in folder "{}":'.format(tabs * '\t', len(all_files_processed),
                                                                              file_suffix, dir_path))
            print('{}\t{}'.format(tabs * '\t', all_files_base))
    else:
        all_files_processed = None
        exception_error(NOT_FOUND.format(dir_path), real_exception=False, tabs=tabs)
    return all_files_processed


def move_file(file_src: str, file_dst: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param file_src:
    :param file_dst:
    :param ack:
    :param tabs:
    :return:
    see move_file_test()
    """
    if os.path.exists(file_src):
        size_s = file_or_folder_size(file_src)
        if os.path.exists(os.path.dirname(file_dst)):
            shutil.move(file_src, file_dst)
            # os.rename(file_src, file_dst)
            # os.replace(file_src, file_dst)
            if ack:
                file_dst_msg = '{}({})'.format(file_dst, size_s)
                print('{}{}'.format(tabs * '\t', MOVED.format(file_src, file_dst_msg)))
        else:
            exception_error(NOT_FOUND.format(os.path.dirname(file_dst)), real_exception=False, tabs=tabs)
    else:
        exception_error(NOT_FOUND.format(file_src), real_exception=False, tabs=tabs)
    return


def copy_file(file_src: str, file_dst: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param file_src:
    :param file_dst:
    :param ack:
    :param tabs:
    :return:
    see copy_file_test()
    """
    if os.path.exists(file_src):  # origin file exists
        size_s = file_or_folder_size(file_src)  # target folder exists
        if os.path.exists(os.path.dirname(file_dst)):  # target file doesn't exists
            if not os.path.exists(file_dst):
                shutil.copyfile(file_src, file_dst)
                if ack:
                    file_dst_msg = '{}({})'.format(file_dst, size_s)
                    print('{}{}'.format(tabs * '\t', COPIED.format(file_src, file_dst_msg)))
            else:
                exception_error(EXISTS.format(file_dst), real_exception=False, tabs=tabs)
        else:
            exception_error(NOT_FOUND.format(os.path.dirname(file_dst)), real_exception=False, tabs=tabs)
    else:
        exception_error(NOT_FOUND.format(file_src), real_exception=False, tabs=tabs)
    return


def delete_file(file: str, ack: bool = True, tabs: int = 1) -> None:
    """
    :param file:
    :param ack:
    :param tabs:
    see delete_file_test()
    """
    if os.path.exists(file):
        size_s = file_or_folder_size(file)
        os.remove(file)
        if ack:
            file_msg = '{}({})'.format(file, size_s)
            print('{}{}'.format(tabs * '\t', DELETED.format(file_msg)))
    else:
        exception_error(NOT_FOUND.format(file), real_exception=False, tabs=tabs)
    return


def delete_files(files: list, ack: bool = True, tabs: int = 1) -> None:
    """
    :param files:
    :param ack:
    :param tabs:
    :return:
    see delete_files_test()
    """
    size_s = 0
    for file in files:
        if os.path.exists(file):
            size_s += file_or_folder_size(file, as_str=False)
            delete_file(file, ack=False, tabs=tabs)
    if ack:
        file_msg = '{}({})'.format(files, convert_size(size_s))
        print('{}{}'.format(tabs * '\t', DELETED.format(file_msg)))
    return


def sleep(seconds: (int, float), ack: bool = False, tabs: int = 1):
    """
    :param seconds:
    :param ack:
    :param tabs:
    :return:
    see sleep_test()
    """
    if ack:
        print('{}Sleeping {} seconds'.format(tabs * '\t', seconds))
    time.sleep(seconds)
    return


def reverse_tuple_or_list(orig: [tuple, list]) -> [tuple, list]:
    """
    :param orig: list or tuple
    :return: dst: reversed list or tuple
    see reverse_tuple_or_list_test()
    """
    dst = orig[::-1]
    return dst


def round_tuple(t: tuple, fp: int = 3, warn: bool = False):
    """
    :param t: tuple 1D of floats
    :param fp: float precision >=0
    :param warn: if to should warning in case of rounding failure(e.g. string in list)
    :return: new_t: rounded tuple
    """
    if fp >= 0:
        new_t = tuple(round_list(li=t, fp=fp, warn=warn))
    else:
        new_t = t
    return new_t


def round_list(li: (list, tuple), fp: int = 3, warn: bool = False):
    """
    :param li: list or tuple 1D of floats
    :param fp: float precision >= 0
    :param warn: if to should warning in case of rounding failure(e.g. string in list)
    :return: new_li: rounded list
    see round_list_test()
    """
    new_li = li
    if fp >= 0:
        try:
            new_np = np.array(li)
            if np.dtype == np.float64 or np.dtype == np.float32 or np.float:
                new_li = np.around(new_np, fp).tolist()
        except TypeError as e:
            if warn:
                exception_error(e, real_exception=True)
    return new_li


def shuffle_np_array(arr: np.array) -> np.array:
    """
    :param arr:
    :return:
    shuffles an array
    see shuffle_np_array_test()
    """
    if is_numpy(arr):
        arr = arr[np.random.permutation(arr.shape[0])]
    return arr


def shuffle_np_arrays(arr_tuple: tuple) -> tuple:
    """
    :param arr_tuple: tuple of arrays (numpy)
        len(arr) is equal on all arrays
    :return: shuffled arrays

    see shuffle_np_arrays_test()
    """
    arrays_size = len(arr_tuple[0])
    rand_perm = np.random.permutation(arrays_size)

    out_tuple = ()
    for arr in arr_tuple:
        arr_shf = arr[rand_perm]
        out_tuple += (arr_shf,)
    return out_tuple


def array_info_print(
        array: np.array,
        title: str = 'var',
        chars: (int, str) = 100,
        fp: (int, None) = 2,
        wm: bool = True,
        rec: bool = False,
        tabs: int = 1
) -> None:
    """
    :param array: (recommended 1d)
    :param title:
    :param chars:
    :param fp:
    :param wm:
    :param rec:
    :param tabs:
    :return:
    prints to_str and then mean, std and sum
    see array_info_print_test()
    """
    print(to_str(var=array, title='{}{}'.format(tabs * '\t', title), chars=chars, fp=fp, wm=wm, rec=rec))
    mean_s = 'mean={:,.%xf}' % fp
    mean_s = mean_s.format(np.mean(array))
    std_s = 'std={:,.%xf}' % fp
    std_s = std_s.format(np.std(array))
    sum_s = 'sum={:,.%xf}' % fp
    sum_s = sum_s.format(np.sum(array))
    print('{}\t{}, {}, {}'.format(tabs * '\t', mean_s, std_s, sum_s))
    return


def get_key_by_value(d: dict, value: any) -> str:
    """
    :param d: dict
    :param value:
    Notice that it will return the first key of the value given. if value not unique...
    :return: the key of the value
    see get_key_by_value_test()
    """
    key = None
    for k, v in d.items():
        if v == value:
            key = k
            break
    return key


def to_hex(num: int) -> str:
    """
    :param num:
    :return:
    convert decimal to hex as str
    see to_hex_and_bin_test()
    """
    string = '{:X}'.format(num)
    return string


def to_bin(num: int) -> str:
    """
    :param num:
    :return:
    convert decimal to bin as str
    see to_hex_and_bin_test()
    """
    string = '{:b}'.format(num)
    return string


def dict_as_table(table: dict, title: str = 'table', fp: int = 2, ack: bool = True, tabs: int = 1) -> str:
    """
    :param table: dict of strings to values. keys must be strings
    :param title:
    :param fp: if v is number - float precision
    :param ack:
    :param tabs:
    :return:
    see dict_as_table_test()
    """
    same_type, type_keys = is_same_type(table.keys())
    string = None
    if same_type and type_keys == 'str':
        str_s = 0  # calculate longest key
        for k in table.keys():
            if len(k) > str_s:
                str_s = len(k)
        string = '{}{}:'.format(tabs * '\t', title)
        s_format_all = '{}\t{:%d}  ==> {}' % str_s
        for i, (k, v) in enumerate(table.items()):
            string += '\n'
            # string += s_format_all.format(tabs * '\t', k, to_str(v, title='key_{}'.format(i), chars=500, fp=fp))
            string += s_format_all.format(tabs * '\t', k, v)
        if ack:
            print(string)
    else:
        exception_error('supports only dict with keys from type str', real_exception=False, tabs=tabs)

    # str_s = 0
    # for k in table.keys():
    #     if len(k) > str_s:
    #         str_s = len(k)
    # s_format_all = '{}\t{:%d}  ==> {}' % str_s
    # # s_format = '{}\t{:%d}  ==> {:%d}' % (str_s, str_s)
    # # s_format_to_num = '{}\t{:%d}  ==> {:%d,}' % (str_s, str_s)
    # # s_format_to_list = '{}\t{:%d}  ==> {}' % str_s
    # for k, v in table.items():
    #     print(s_format_all.format(tabs * '\t', k, to_str(v)))
    #     # if is_int(v) or is_float(v):
    #     #     print(s_format_to_num.format(tabs * '\t', k, round(v, fp)))
    #     # elif is_list(v):
    #     #     print(s_format_to_list.format(tabs * '\t', k, v))
    #     # else:
    #     #     print(s_format.format(tabs * '\t', k, v))
    return string


def is_same_type(container: (list, tuple)) -> (bool, str):
    """
    :param container:
    :return:
    check if all items in list or tuple are of the same type
    if they are, return the type
    see is_same_type_test
    """
    same_type, first_type = True, None
    if len(container) > 0:
        it = iter(container)
        first_type = type(next(it))
        # for item in container:
        #     print(type(item))
        same_type = all((type(x) is first_type) for x in it)
        if not same_type:
            first_type = None
        else:
            first_type = str(first_type).replace('<class \'', '').replace('\'>', '')

    return same_type, first_type


def convert_size(size_bytes: int) -> str:
    """
    :param size_bytes:
    :return:
    get size in bytes and return string e.g. '241.19 MB'
    """
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return "%s %s" % (s, size_name[i])


def hard_disc(one_liner: bool = False, tabs: int = 1):
    """
    :param one_liner: per partition
    :param tabs:
    :return:
    see hard_disc_test()
    """
    partitions = psutil.disk_partitions()

    if one_liner:
        string = '{}'.format(tabs * '\t')
        for i, partition in enumerate(partitions):
            try:
                partition_usage = psutil.disk_usage(partition.mountpoint)
                string_device = '{}: Total {}, Used {}({:.2f}%), Free {}'.format(
                    partition.device.replace(':\\', ''),
                    convert_size(partition_usage.total),
                    convert_size(partition_usage.used),
                    partition_usage.percent,
                    convert_size(partition_usage.free),
                )
            except PermissionError as e:
                string_device = add_color('{}: PermissionError: {}'.format(partition.device.replace(':\\', ''), e))

            string += string_device
            if i < (len(partitions) - 1):
                string += ', '
    else:
        string = '{}disk space:'.format(tabs * '\t')
        for i, partition in enumerate(partitions):
            try:
                partition_usage = psutil.disk_usage(partition.mountpoint)
                string_device = '\n{}\t{}:'.format(tabs * '\t', partition.device.replace(':\\', ''))
                string_device += '\n{}\t\tTotal {}'.format(tabs * '\t', convert_size(partition_usage.total))
                string_device += '\n{}\t\tUsed {}({:.2f}%)'.format(tabs * '\t', partition_usage.used,
                                                                   partition_usage.percent)
                string_device += '\n{}\t\tFree {}'.format(tabs * '\t', convert_size(partition_usage.free))
            except PermissionError as e:
                string_device = '\n{}\t{}:'.format(tabs * '\t', partition.device.replace(':\\', ''))
                string_device += add_color('\n{}\t\tPermissionError: {}'.format(tabs * '\t', e))

            string += string_device
            if i < (len(partitions) - 1):
                string += ', '
    return string


def ram_size(one_liner: bool = False, tabs: int = 1):
    """
    :param one_liner:
    :param tabs:
    :return:
    see ram_size_test()
    """
    mem = psutil.virtual_memory()
    if one_liner:
        string = '{}Total {}, Used {}({}%), Available {} '.format(
            tabs * '\t',
            convert_size(mem.total),
            convert_size(mem.used),
            mem.percent,
            convert_size(mem.available),
        )
    else:
        string = '{}RAM Info:'.format(tabs * '\t')
        string += '\n{}\tTotal {}'.format(tabs * '\t', convert_size(mem.total))
        string += '\n{}\tUsed {}({}%)'.format(tabs * '\t', convert_size(mem.used), mem.percent)
        string += '\n{}\tAvailable {}'.format(tabs * '\t', convert_size(mem.available))
    return string


def cpu_info(one_liner: bool = False, tabs: int = 1):
    """
    :param one_liner:
    :param tabs:
    :return:
    see cpu_info_test()
    """
    if one_liner:
        uname = platform.uname()
        cpufreq = psutil.cpu_freq()
        string = '{}'.format(tabs * '\t')
        string += '{}, '.format(uname.machine)
        string += '{}, '.format(uname.processor)
        string += 'Physical cores {}, '.format(psutil.cpu_count(logical=False))
        string += 'Total cores {}, '.format(psutil.cpu_count(logical=True))
        string += 'Frequency {:.2f}Mhz, '.format(cpufreq.current)
        string += 'CPU Usage {}%'.format(psutil.cpu_percent())
    else:
        uname = platform.uname()
        cpufreq = psutil.cpu_freq()
        string = '{}System Info:'.format(tabs * '\t')
        string += '\n{}\tSystem {}'.format(tabs * '\t', uname.system)
        string += '\n{}\tName {}'.format(tabs * '\t', uname.node)
        string += '\n{}\tRelease {}'.format(tabs * '\t', uname.release)
        string += '\n{}\tVersion {}'.format(tabs * '\t', uname.version)
        string += '\n{}\tMachine {}'.format(tabs * '\t', uname.machine)
        string += '\n{}\tProcessor {}'.format(tabs * '\t', uname.processor)
        string += '\n{}\tPhysical cores {}'.format(tabs * '\t', (psutil.cpu_count(logical=False)))
        string += '\n{}\tTotal cores {}'.format(tabs * '\t', (psutil.cpu_count(logical=True)))
        string += '\n{}\tMax Frequency {:.2f}Mhz'.format(tabs * '\t', cpufreq.max)
        string += '\n{}\tMin Frequency {:.2f}Mhz'.format(tabs * '\t', cpufreq.min)
        string += '\n{}\tCurrent Frequency {:.2f}Mhz'.format(tabs * '\t', cpufreq.current)
        string += '\n{}\tCPU Usage {}%'.format(tabs * '\t', psutil.cpu_percent())
        per_core = ''
        for i, percentage in enumerate(psutil.cpu_percent(percpu=True, interval=1)):
            # print(f"Core {i}: {percentage}%")
            per_core += 'C{}-{}% '.format(i, percentage)
        string += '\n{}\tCPU Usage per core {}'.format(tabs * '\t', per_core)
    return string


def wizzi_utils_requirements():
    pip = 'pip install {}'
    pack = 'Package {}({}.py, namespace: {}):'
    pack_test = '\ttest_{}.py(namespace: {}.test):'
    sub_pack = 'Sub Package: {}'

    MT_PACK, MT_MOD, MT_SC = 'misc', 'misc_tools', 'mt'
    ALG_PACK, ALG_MOD, ALG_SC = 'algorithms', 'algorithms', 'alg'
    COT_PACK, COT_MOD, COT_SC = 'coreset', 'coreset_tools', 'cot'
    GOT_PACK, GOT_MOD, GOT_SC = 'google', 'google_tools', 'got'
    JT_PACK, JT_MOD, JT_SC = 'json', 'json_tools', 'jt'
    CVT_PACK, CVT_MOD, CVT_SC = 'open_cv', 'open_cv_tools', 'cvt'
    PYPLT_PACK, PYPLT_MOD, PYPLT_SC = 'pyplot', 'pyplot_tools', 'pyplt'
    ST_PACK, ST_MOD, ST_SC = 'socket', 'socket_tools', 'st'
    TT_PACK, TT_MOD, TT_SC = 'torch', 'torch_tools', 'tt'
    TFT_PACK, TFT_MOD, TFT_SC = 'tensorflow', 'tensorflow_tools', 'tft'
    TFLT_PACK, TFLT_MOD, TFLT_SC = 'tflite', 'tflite_tools', 'tflt'
    TTS_PACK, TTS_MOD, TTS_SC = 'tts', 'tts', 'tts'
    MODELS_PACK, MODELS_MOD, MODELS_SC = 'models', 'models', 'models'
    print('Requirements per package:')
    print('\tTo use a test of a package, must install all package requirements')

    print(LINES)
    print(pack.format(MT_PACK, MT_MOD, MT_SC))
    print('\t{} # to use 7z compress and extract'.format(pip.format('py7zr')))
    print(pack_test.format(MT_MOD, MT_SC))
    print('\t\tNone')

    print(LINES)
    print(pack.format(ALG_PACK, ALG_MOD, ALG_SC))
    print('\t{}'.format(pip.format('sklrean')))
    print(pack_test.format(ALG_MOD, ALG_SC))
    print('\t\t{}'.format(sub_pack.format(PYPLT_PACK)))

    print(LINES)
    print(pack.format(COT_PACK, COT_MOD, COT_SC))
    print('\tNone')
    print(pack_test.format(COT_MOD, COT_SC))
    print('\t\tNone')

    print(LINES)
    print(pack.format(GOT_PACK, GOT_MOD, GOT_SC))
    print('\t*** see instructions in order to use written in class google_handler docstring')
    print('\t{}'.format(pip.format('pydrive')))
    print('\t{}'.format(pip.format('yaml')))
    print(pack_test.format(GOT_MOD, GOT_SC))
    print('\t\t{}'.format(sub_pack.format(ST_PACK)))

    print(LINES)
    print(pack.format(JT_PACK, JT_MOD, JT_SC))
    print('\tNone')
    print(pack_test.format(JT_MOD, JT_SC))
    print('\t\tNone')

    print(LINES)
    print(pack.format(CVT_PACK, CVT_MOD, CVT_SC))
    print('\t{}'.format(pip.format('opencv-python')))
    print('\t{}'.format(sub_pack.format(PYPLT_PACK)))
    print('\t{}'.format(sub_pack.format(ST_PACK)))
    print('\t{} # for CameraWu type_cam=="imutils"'.format(pip.format('imutils')))
    print('\t{} # for CameraWu type_cam=="acapture"'.format(pip.format('acapture')))
    print(pack_test.format(CVT_MOD, CVT_SC))
    print('\t\t{}'.format(sub_pack.format(ST_PACK)))
    print('\t\t{}'.format(sub_pack.format(CVT_PACK)))

    print(LINES)
    print(pack.format(PYPLT_PACK, PYPLT_MOD, PYPLT_SC))
    print('\t{}'.format(pip.format('matplotlib')))
    print('\t{}'.format(pip.format('mpl_toolkits')))
    print('\t{} # for compare_images_multi_sets_squeezed()'.format(pip.format('torch')))
    print(pack_test.format(PYPLT_MOD, PYPLT_SC))
    print('\t\t{} # for compare_images_multi_sets_squeezed_test()'.format(pip.format('torch')))
    print('\t\t{} # also the test'.format(sub_pack.format(CVT_PACK)))

    print(LINES)
    print(pack.format(ST_PACK, ST_MOD, ST_SC))
    print('\tNone')
    print(pack_test.format(ST_MOD, ST_SC))
    print('\t\t{}'.format(sub_pack.format(JT_PACK)))

    print(LINES)
    url = 'https://download.pytorch.org/whl/cu90/torch_stable.html'
    print(pack.format(TT_PACK, TT_MOD, TT_SC))
    print('\ttorch options:')
    print('\t\t{} # cpu'.format(pip.format('torch')))
    print('\t\t{} # gpu(cuda 9 and 9.1)'.format(pip.format('torch==1.1.0 torchvision==0.3.0 -f {}'.format(url))))
    print('\t\tall options: https://pytorch.org/get-started/locally/')
    print('\t\t** may need to uninstall numpy and re-install')
    print('\t{}'.format(pip.format('torchvision')))
    print(pack_test.format(TT_MOD, TT_SC))
    print('\t\tNone')

    print(LINES)
    print(pack.format(TFT_PACK, TFT_MOD, TFT_SC))
    print('\ttensorflow options:')
    print('\t\t{} # cpu only'.format(pip.format('tensorflow')))
    print('\t\t{} # gpu cuda 9.1'.format(pip.format('tensorflow-gpu==1.12')))
    print('\t\t{} # gpu cuda>=10'.format(pip.format('tensorflow-gpu')))
    url = 'https://stackoverflow.com/questions/50622525/'
    url += 'which-tensorflow-and-cuda-version-combinations-are-compatible'
    print('\t\tall options: {}'.format(url))
    print(pack_test.format(TFT_MOD, TFT_SC))
    print('\t\tNone')

    print(LINES)
    print(pack.format(TFLT_PACK, TFLT_MOD, TFLT_SC))
    print('\ttflite options:')
    url = 'https://google-coral.github.io/py-repo/'
    print('\t\t{} # windows 10'.format(pip.format('--extra-index-url {} tflite_runtime'.format(url))))
    print(pack_test.format(TFLT_MOD, TFLT_SC))
    print('\t\t{}'.format(None))

    print(LINES)
    print(pack.format(TTS_PACK, TTS_MOD, TTS_SC))
    print('\t{}'.format(pip.format('pyttsx3')))
    print(pack_test.format(TTS_MOD, TTS_SC))
    print('\t\tNone')

    print(LINES)
    print(pack.format(MODELS_PACK, MODELS_MOD, MODELS_SC))
    print('\t{}'.format(sub_pack.format(ST_PACK)))
    print('\t{}'.format(sub_pack.format(PYPLT_PACK)))
    print('\t{}'.format(sub_pack.format(CVT_PACK)))
    print(pack_test.format(MODELS_MOD, MODELS_SC))
    print('\t\tNone')
    print(LINES)
    return


def last_exception(tabs: int = 1):
    """
    prints last_exception occurred
    :return:
    """
    exception_error(sys.exc_info()[1], tabs=tabs)
    return


def file_or_folder_size(path: str, as_str: bool = True, ack: bool = False, tabs: int = 1) -> (int, str):
    """
    :param path:
    :param as_str:
    :param ack:
    :param tabs:
    :return: file or folder size
      as_str True: as str - see convert_size()
      as_str False: in bytes
    see file_or_folder_size_test()
    """
    size_s = 0
    if os.path.exists(path):
        if os.path.isfile(path):
            size_s = os.stat(path).st_size
            prefix = 'file'
        else:
            for _dir_path, dir_names, file_names in os.walk(path):
                for f in file_names:
                    fp = os.path.join(_dir_path, f)
                    if not os.path.islink(fp):  # not symbolic link
                        size_s += os.path.getsize(fp)
            # size_s = sum(d.stat().st_size for d in os.scandir(path) if d.is_file())
            prefix = 'folder'

        if as_str:
            size_s = convert_size(size_s)
        if ack:
            print('{}{} {} size = {}'.format(tabs * '\t', prefix, path, size_s))
    else:
        exception_error(NOT_FOUND.format(path), real_exception=False, tabs=tabs)
    return size_s


def full_path_no_limit(path_file_or_dir: str) -> str:
    """
    without this basename can't exceed 70 chars
    with    this basename can't exceed 255 chars
    return full path with no size limit on the length
    :param path_file_or_dir:
    :return:
    """
    path_file_or_dir = "\\\\?\\" + os.path.abspath(path_file_or_dir)
    return path_file_or_dir


def extract_file(
        src: str,
        dst_folder: str = 'ex',
        file_type: str = 'zip',
        ack: bool = True,
        tabs: int = 1):
    """
    :param src: compressed file (zip, 7z, tar)
    :param dst_folder: destination folder.
        if folder not empty and in the zip there is a file that exists in the folder, it will be overwritten
    :param file_type: zip, 7z, tar
    :param ack:
    :param tabs:
    :return:
    see compress_and_extract_test()
    """
    if os.path.exists(src):
        if file_type in ['zip', 'tar', '7z']:
            if file_type == '7z':
                try:
                    if ('7z', ['.7z'], '7zip archive') not in shutil.get_unpack_formats():
                        # noinspection PyPackageRequirements,PyUnresolvedReferences
                        from py7zr import unpack_7zarchive
                        shutil.register_unpack_format(name='7z', extensions=['.7z'], function=unpack_7zarchive,
                                                      description='7zip archive')
                except (ModuleNotFoundError, AttributeError) as e:
                    exception_error(e, real_exception=True, tabs=tabs)
                    return
            shutil.unpack_archive(filename=src, extract_dir=dst_folder)

            if ack:
                msg = '{}extracted {}({}) to {}({})'.format(
                    tabs * '\t',
                    src, file_or_folder_size(src),
                    dst_folder, file_or_folder_size(dst_folder)
                )
                print(add_color(msg, ops=SUCCESS_C))
    else:
        exception_error(NOT_FOUND.format(src), real_exception=False, tabs=tabs)
    return


def compress_file_or_folder(
        src: str,
        dst_path: str = './file',
        file_type: str = 'zip',
        ack: bool = True,
        tabs: int = 1) -> None:
    """
    :param src: folder or file to compress
    :param dst_path: the compressed file name (no suffix)
        e.g. if file_type=='zip':  dst_path=./a -> ./a.zip
    :param file_type: 'zip', 'tar', '7z'
    :param ack:
    :param tabs:
    :return:
    see compress_and_extract_test()
    """
    if os.path.exists(src):
        if file_type in ['zip', 'tar', '7z']:
            if file_type == '7z':
                try:
                    if ('7z', '7zip archive') not in shutil.get_archive_formats():
                        # noinspection PyPackageRequirements,PyUnresolvedReferences
                        from py7zr import pack_7zarchive
                        shutil.register_archive_format(name='7z', function=pack_7zarchive, description='7zip archive')
                except (ModuleNotFoundError, AttributeError) as e:
                    exception_error(e, real_exception=True, tabs=tabs)
                    return
            if os.path.isfile(src):
                src_file = os.path.basename(src)
                src_dir = os.path.dirname(src)
                shutil.make_archive(base_name=dst_path, format=file_type, root_dir=src_dir, base_dir=src_file)
            else:
                shutil.make_archive(base_name=dst_path, format=file_type, root_dir=src)

            if ack:
                dst_path += '.{}'.format(file_type)
                msg = '{}compressed {}({}) to {}({})'.format(
                    tabs * '\t',
                    src, file_or_folder_size(src),
                    dst_path, file_or_folder_size(dst_path)
                )
                print(add_color(msg, ops=SUCCESS_C))
    else:
        exception_error(NOT_FOUND.format(src), real_exception=False, tabs=tabs)
    return


class FPS:
    """
    measure times easily
    see classFPS_test()
    """

    def __init__(self, summary_title: str = None) -> None:
        self.timer_begin = None
        self.timer_end = None
        self.times_elapsed = []
        self.summary_title = summary_title
        return

    def __del__(self) -> None:
        return

    def __str__(self) -> str:
        return ''

    def start(self, ack_progress: bool = False, tabs: int = 1, with_title: bool = False) -> None:
        if ack_progress:
            string = tabs * '\t'
            if with_title:
                string += '{}:'.format(self.summary_title)
            string += 'Iter {} Started:'.format(len(self.times_elapsed) + 1)
            print(string)
        self.timer_begin = get_timer()
        self.timer_end = None
        return

    def update(self, ack_progress: bool = False, tabs: int = 1, with_title: bool = False) -> None:
        if self.timer_begin is None:
            exception_error(e='Must call FPS::start before update', real_exception=False, tabs=0)
        else:
            self.timer_end = get_timer()
            time_elapsed = self.timer_end - self.timer_begin
            self.times_elapsed.append(time_elapsed)
            if ack_progress:
                self.so_far(tabs, with_title)
            self.timer_begin = None
            self.timer_end = None
        return

    # def convert(self) -> str:
    #     rt = self.times_elapsed[-1]
    #     if rt < 0.001:  # 0.0001 -> 0.100ms
    #         rt_str = '{:.3f}ms'.format(rt * 1000)  # fp=6 digit
    #     elif rt < 0.1:
    #         rt_str = '{:.2f}ms'.format(rt * 1000)  # fp=6 digit
    #     elif rt < 1:
    #         rt_str = '{:.1f}ms'.format(rt * 1000)
    #     elif rt < 20:
    #         rt_str = '{:.3f}s'.format(rt)
    #     elif rt < 60:
    #         rt_str = '{:.1f}s'.format(rt)
    #     elif rt < 120:
    #         m, s = divmod(rt, 60)
    #         h, m = divmod(m, 60)
    #         rt_str = '{}:{:.1f}'.format(int(m), s)
    #     print('\trt={:.6f}s == {}'.format(rt, rt_str))
    #     return rt_str

    def so_far(self, tabs: int = 1, with_title: bool = False) -> None:
        if len(self.times_elapsed) > 0:
            string = tabs * '\t'
            if with_title:
                string += '{}:'.format(self.summary_title)
            string += 'Iter {} Done: '.format(len(self.times_elapsed))

            # rt_str = self.convert()
            rt_str = '{:.3f}s'.format(self.times_elapsed[-1])

            string += 'iterTime={}, '.format(rt_str)
            avg_times = np.mean(self.times_elapsed)
            if len(self.times_elapsed) > 1:
                string += 'totalTime={:.3f}s, '.format(np.sum(self.times_elapsed))
                string += 'avgTime={:.3f}s, '.format(avg_times)
            if avg_times > 0.0001:
                fps = '{:.2f} FPS'.format(1 / avg_times)
            else:
                fps = 'FPS very high'.format(tabs * '\t')
            string += add_color(fps, ops='r')
            print(string)
        return

    def finalize(self, tabs: int = 1) -> None:
        tabs_s = tabs * '\t'
        if len(self.times_elapsed) <= 0:
            exception_error('Times_elapsed is empty', real_exception=False, tabs=tabs)
        else:
            # title
            title = 'Summary '
            if self.summary_title is not None:
                title += 'of {}'.format(self.summary_title)
            # add rounds to title
            title += '({} iters)'.format(len(self.times_elapsed))
            print('{}{}'.format(tabs_s, add_color(title, ops='underlined')))

            # total time
            total = '{}\tTotal   Run Time = {:.3f}s'.format(tabs_s, np.sum(self.times_elapsed))
            print(total)

            # average time
            avg_times = np.mean(self.times_elapsed)
            std_times = np.std(self.times_elapsed)
            avg_str = '{}\tAverage Run Time = {:.3f}s (std = {:.3f})'.format(tabs_s, avg_times, std_times)
            print(avg_str)
            # FPS
            if avg_times > 0.0001:
                fps = '{}\t{:.2f} FPS'.format(tabs * '\t', 1 / avg_times)
            else:
                fps = '{}\tFPS very high'.format(tabs * '\t')
            print(add_color(fps, ops='r'))
        return



from wizzi_utils.tflite import tflite_tools as tflt
from wizzi_utils.misc import misc_tools as mt


def get_tflite_version_test():
    mt.get_function_name(ack=True, tabs=0)
    tflt.get_tflite_version(ack=True)
    return


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))
    get_tflite_version_test()
    print('{}'.format('-' * 20))
    return



import numpy as np
import os
import abc
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.socket import socket_tools as st
from wizzi_utils.models import models_configs as cfg


class BaseModel:
    def __init__(self, save_load_dir: str, model_name: str):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name from models_config.py
        see:
        class_Models_test()
        """
        self.model_name = model_name
        self.local_path = save_load_dir
        if not os.path.exists(save_load_dir):
            mt.create_dir(save_load_dir, ack=False)
        return

    @staticmethod
    def model_name_valid(model_name: str) -> bool:
        if model_name not in cfg.MODELS_CONFIG:
            err = 'model {} not found in MODELS_CONFIG. options: {}'
            mt.exception_error(err.format(model_name, list(cfg.MODELS_CONFIG.keys())), real_exception=False)
            valid = False
        else:
            valid = True
        return valid

    @staticmethod
    def model_type_valid(model_type: str, model_type_needed: str) -> bool:
        if model_type != model_type_needed:
            err = 'model type is {} but {} needed'.format(model_type, model_type_needed)
            mt.exception_error(err, real_exception=False)
            valid = False
        else:
            valid = True
        return valid

    @staticmethod
    def _download_if_needed(local_path: str, url_dict: dict) -> None:
        if not os.path.exists(local_path):
            download_style = url_dict['download_style']
            url = url_dict['url']
            if download_style == cfg.DownloadStyle.Direct.value:
                st.download_file(url=url, dst_path=local_path)
            elif download_style in [cfg.DownloadStyle.Tar.value, cfg.DownloadStyle.Zip.value]:
                root_d = os.path.dirname(local_path)
                comp_path = '{}/compressed.{}'.format(root_d, download_style)
                st.download_file(url=url, dst_path=comp_path)  # download compressed
                extracted_folder = '{}/ex'.format(root_d)
                if mt.is_windows():
                    extracted_folder = mt.full_path_no_limit(extracted_folder)

                mt.extract_file(src=comp_path, dst_folder=extracted_folder,
                                file_type=download_style)  # extract
                target_file_in_tar = url_dict['file_to_look']
                # find target file
                target_files = mt.find_files_in_folder(dir_path=extracted_folder, file_suffix=target_file_in_tar)
                if len(target_files) != 1:
                    err_msg = 'not found or found more than 1 target file in downloaded folder: {}'.format(target_files)
                    mt.exception_error(err_msg, real_exception=False)
                    exit(-1)
                mt.move_file(file_src=target_files[0], file_dst=local_path)  # move file

                mt.delete_file(file=comp_path)  # clean up tar\zip
                mt.delete_dir_with_files(dir_path=extracted_folder)  # clean up extracted_folder
        return

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        print('abs method - needs implementation')
        exit(-1)
        return []



"""
:requires: tflite_runtime
tf_lite install: https://www.tensorflow.org/lite/guide/python#install_tensorflow_lite_for_python
    winOs: pip install --extra-index-url https://google-coral.github.io/py-repo/ tflite_runtime
"""
try:
    from wizzi_utils.tflite.tflite_tools import *
except ModuleNotFoundError as e:
    pass

from wizzi_utils.tflite import test



import os
import json
from wizzi_utils.misc import misc_tools as mt


def json_to_string(j: dict, indent: int = -1, sort_keys: bool = False, tabs: int = 0) -> str:
    """
    :param j: dict
    :param indent: how many indents
    :param sort_keys: sort dict keys
    :param tabs:
    :return: string rep of j
    see json_to_string_test()
    """
    if indent == -1:
        indent = None
    string = json.dumps(j, indent=indent, sort_keys=sort_keys)
    if tabs > 0:
        string = '\t' * tabs + string
        string = string.replace('\n', '\n{}'.format(tabs * '\t'))
    return string


def string_to_json(j_str: str) -> json:
    """
    changes a string to a json dict
    def string_to_json_test():
    """
    return json.loads(j_str)


def load_json(file_path: str, ack: bool = True, tabs: int = 1) -> dict:
    """
    loads a dict in json format from path
    see save_load_json_test()
    """
    ret_dict = {}
    if os.path.exists(file_path):
        ret_dict = json.load(open(file_path))
        if ack:
            size_s = mt.file_or_folder_size(file_path)
            file_msg = '{}({})'.format(file_path, size_s)
            print('{}{}. {}'.format(tabs * '\t', mt.LOADED.format(file_msg),
                                    mt.CONTENT.format(json_to_string(ret_dict))))
    else:
        mt.exception_error(mt.NOT_FOUND.format(file_path), real_exception=False, tabs=tabs)
    return ret_dict


def load_jsons(files_path: list, ack: bool = True, tabs: int = 1) -> dict:
    """
    loads several of json files format from paths and concat to one dict
    asserts if a key found on 2 of the files
    see save_load_json_test()
    """
    all_in_one_dict = {}
    len_keys_json = 0
    size_s = 0
    for file_path in files_path:
        size_s += mt.file_or_folder_size(file_path, as_str=False)
        j = load_json(file_path, ack=False, tabs=tabs)
        len_keys_json += len(j)
        all_in_one_dict.update(j)

    assert len_keys_json == len(all_in_one_dict), 'Duplicated keys found: {}'.format(files_path)
    if ack:
        file_msg = '{}({})'.format(files_path, mt.convert_size(size_s))
        print('{}{}. {}'.format(tabs * '\t', mt.LOADED.format(file_msg),
                                mt.CONTENT.format(json_to_string(all_in_one_dict))))
    return all_in_one_dict


def save_json(file_path: str,
              j: dict,
              indent: int = -1,
              sort_keys: bool = False,
              ack: bool = True,
              tabs: int = 1
              ) -> None:
    """
    see save_load_json_test()
    """
    json.dump(j, open(file_path, 'w'), indent=indent, sort_keys=sort_keys)
    if ack:
        size_s = mt.file_or_folder_size(file_path)
        file_msg = '{}({})'.format(file_path, size_s)
        print('{}{}. {}'.format(tabs * '\t', mt.SAVED.format(file_msg),
                                mt.CONTENT.format(json_to_string(j, indent=indent, sort_keys=sort_keys))))
    return



import numpy as np
import math
import os
from wizzi_utils.pyplot import pyplot_tools as pyplt
from wizzi_utils.misc import misc_tools as mt
# noinspection PyPackageRequirements
import cv2


def get_cv_version(ack: bool = False, tabs: int = 1) -> str:
    """ see get_cv_version_test() """
    """ see get_cv_version_test() """
    string = mt.add_color('{}* OpenCv Version {}'.format(tabs * '\t', cv2.getVersionString()), ops=mt.SUCCESS_C)
    if ack:
        print(string)
    return string


def load_img(path: str, ack: bool = True, tabs: int = 1) -> np.array:
    """ see imread_imwrite_test() """
    if os.path.exists(path):
        img = cv2.imread(path)
        if ack:
            size_s = mt.file_or_folder_size(path)
            file_msg = '{}({}) of shape {}'.format(path, size_s, img.shape)
            print('{}{}'.format(tabs * '\t', mt.LOADED.format(file_msg)))
    else:
        mt.exception_error(mt.NOT_FOUND.format(path), real_exception=False, tabs=tabs)
        img = None
    return img


def save_img(path: str, img: np.array, ack: bool = True, tabs: int = 1) -> None:
    """ see imread_imwrite_test() """
    if os.path.exists(os.path.dirname(path)):
        cv2.imwrite(path, img)
        if ack:
            size_s = mt.file_or_folder_size(path)
            file_msg = '{}({}) of shape {}'.format(path, size_s, img.shape)
            print('{}{}'.format(tabs * '\t', mt.SAVED.format(file_msg)))
    else:
        mt.exception_error(mt.NOT_FOUND.format(os.path.dirname(path)), real_exception=False)
    return


def list_to_cv_image(cv_img: [list, np.array]) -> np.array:
    """
    :param cv_img: numpy or list. if list: convert to numpy with dtype uint8
    :return: cv_img
    see list_to_cv_image_test()
    """
    if mt.is_list(cv_img):
        cv_img = np.array(cv_img, dtype='uint8')
    return cv_img


def display_open_cv_image(
        img: np.array,
        ms: int = 0,
        title: str = 'cv_image',
        loc: (tuple, str) = None,
        resize: float = None,
        header: str = None,
        save_path: str = None,
        tabs: int = 1,
) -> None:
    """
    :param img: cv image in numpy array or list
    :param ms: 0 blocks, else time in milliseconds before image is closed
    :param title: window title
    :param loc: top left corner window location
        if tuple: x,y coordinates
        if str: see Location enum in pyplt. e.g. pyplt.Location.TOP_LEFT.value (which is 'top_left')
    :param resize: None for original size. else img_size *= resize (resize > 0)
    :param header: text to add at the top left
    :param save_path: if not none, saves image to this path
    :param tabs:
    see display_open_cv_image_test()
    """
    img = list_to_cv_image(img)
    if resize is not None:
        img = resize_opencv_image(img, resize)
    if header is not None:
        add_header(img, header, bg_font_scale=2)
    if save_path is not None:
        save_img(path=save_path, img=img, ack=True, tabs=tabs)
    cv2.imshow(title, img)
    if loc is not None:
        if mt.is_str(loc):
            move_cv_img_by_str(img, title, where=loc)
        elif mt.is_tuple(loc):
            move_cv_img_x_y(title, x_y=loc)
    cv2.waitKey(ms)
    return


def resize_opencv_image(img: np.array, scale_percent: float):
    """
    :param img: cv img
    :param scale_percent: float>0(could be bigger than 1)
    img_size *= scale_percent
    see resize_opencv_image_test()
    """
    if scale_percent is None or scale_percent <= 0:
        resize_image = img
        mt.exception_error('illegal value for scale_percent={}'.format(scale_percent), real_exception=False, tabs=0)
    else:
        width = math.ceil(img.shape[1] * scale_percent)
        height = math.ceil(img.shape[0] * scale_percent)
        dim = (width, height)
        resize_image = cv2.resize(img, dim, interpolation=cv2.INTER_AREA)
    return resize_image


def move_cv_img_x_y(win_title: str, x_y: tuple) -> None:
    """
    :param win_title: cv img with win_title to be moved
    :param x_y: tuple of ints. x,y of top left corner
    Move cv img upper left corner to pixel (x, y)
    see move_cv_img_x_y_test()
    """
    cv2.moveWindow(win_title, x_y[0], x_y[1])
    return


def move_cv_img_by_str(img: np.array, win_title: str, where: str = pyplt.Location.TOP_LEFT.value) -> None:
    """
    :param img: to get the dims of the image
    :param win_title: cv img with win_title to be moved
    :param where: see Location enum in pyplt. e.g. pyplt.Location.TOP_LEFT.value (which is 'top_left')
    Move cv img upper left corner to pixel (x, y)
    see move_cv_img_by_str_test()
    """
    try:
        window_w, window_h = pyplt.screen_dims()  # screen dims in pixels
        if window_w != -1 and window_h != -1:
            fig_w, fig_h = img.shape[1], img.shape[0]
            task_bar_offset = 75  # about 75 pixels due to task bar

            x, y = 0, 0  # Location.TOP_LEFT.value: default
            if where == pyplt.Location.TOP_CENTER.value:
                x = (window_w - fig_w) / 2
                y = 0
            elif where == pyplt.Location.TOP_RIGHT.value:
                x = window_w - fig_w
                y = 0
            elif where == pyplt.Location.CENTER_LEFT.value:
                x = 0
                y = (window_h - fig_h - task_bar_offset) / 2
            elif where == pyplt.Location.CENTER_CENTER.value:
                x = (window_w - fig_w) / 2
                y = (window_h - fig_h - task_bar_offset) / 2
            elif where == pyplt.Location.CENTER_RIGHT.value:
                x = window_w - fig_w
                y = (window_h - fig_h - task_bar_offset) / 2
            elif where == pyplt.Location.BOTTOM_LEFT.value:
                x = 0
                y = window_h - fig_h - task_bar_offset
            elif where == pyplt.Location.BOTTOM_CENTER.value:
                x = (window_w - fig_w) / 2
                y = window_h - fig_h - task_bar_offset
            elif where == pyplt.Location.BOTTOM_RIGHT.value:
                x = window_w - fig_w
                y = window_h - fig_h - task_bar_offset
            x, y = int(x), int(y)
            move_cv_img_x_y(win_title, x_y=(x, y))
        else:
            move_cv_img_x_y(win_title, x_y=(0, 0))
    except (ValueError, Exception) as e:
        mt.exception_error(e, real_exception=True)
        move_cv_img_x_y(win_title, x_y=(0, 0))
    return


def unpack_list_imgs_to_big_image(imgs: list, grid: tuple, resize: float = None) -> np.array:
    """
    :param imgs: list of cv images
    :param resize: resize factor
    :param grid: the layout you want as output
        (1,len(imgs)): 1 row
        (len(imgs),1): 1 col
        (2,2): 2x2 grid - supports len(imgs)<=4 but not more
    see unpack_list_imgs_to_big_image_test()
    """
    for i in range(len(imgs)):
        imgs[i] = list_to_cv_image(imgs[i])
        if resize is not None:
            imgs[i] = resize_opencv_image(imgs[i], resize)
        if len(imgs[i].shape) == 2:  # if gray - see as rgb
            imgs[i] = gray_scale_img_to_BGR_form(imgs[i])

    imgs_n = len(imgs)
    if imgs_n == 1:
        big_img = imgs[0]
    else:
        padding_bgr = list(pyplt.get_BGR_color('red'))
        height, width, cnls = imgs[0].shape
        rows, cols = grid
        big_img = np.zeros(shape=(height * rows, width * cols, cnls), dtype='uint8') + 255  # white big image

        row_ind, col_ind = 1, 1
        for i, img in enumerate(imgs):
            h_begin, h_end = height * (row_ind - 1), height * row_ind
            w_begin, w_end = width * (col_ind - 1), width * col_ind
            big_img[h_begin:h_end, w_begin:w_end, :] = img  # 0

            if rows > 1:  # draw bounding box on the edges. no need if there is 1 row or 1 col
                big_img[h_begin, w_begin:w_end, :] = padding_bgr
                big_img[h_end - 1, w_begin:w_end - 1, :] = padding_bgr
            if cols > 1:
                big_img[h_begin:h_end, w_begin, :] = padding_bgr
                big_img[h_begin:h_end, w_end - 1, :] = padding_bgr

            col_ind += 1
            if col_ind > cols:
                col_ind = 1
                row_ind += 1
    return big_img


def display_open_cv_images(
        imgs: list,
        ms: int = 0,
        title: str = 'cv_image',
        loc: (tuple, str) = None,
        resize: float = None,
        grid: tuple = (1, 2),
        header: str = None,
        save_path: str = None,
        tabs: int = 1,
) -> None:
    """
    :param imgs: list of RGB or gray scale images
    :param ms: 0 blocks, else time in milliseconds before image is closed
    :param title: window title
    :param loc: top left corner window location
        if tuple: x,y coordinates
        if str: see Location enum in pyplt. e.g. pyplt.Location.TOP_LEFT.value (which is 'top_left')
    :param resize: None for original size. else img_size *= resize (resize > 0)
    :param grid: size of rows and cols of the new image. e.g. (2,1) 2 rows with 1 img on each
        grid slots must be >= len(imgs)
    :param header: text to add at the top left
    :param save_path: if not none, saves image to this path
    :param tabs:
    see display_open_cv_images_test()
    """
    imgs_n = len(imgs)
    if imgs_n > 0:
        total_slots = grid[0] * grid[1]
        assert imgs_n <= total_slots, 'grid has {} total_slots, but len(imgs)={}'.format(total_slots, imgs_n)
        big_img = unpack_list_imgs_to_big_image(imgs, grid=grid, resize=resize)
        display_open_cv_image(big_img, ms=ms, title=title, loc=loc, resize=None, header=header, save_path=save_path,
                              tabs=tabs)
    return


def gray_scale_img_to_BGR_form(gray_img: np.array) -> np.array:
    """
    :param gray_img: from shape (x,y) - 1 channel (gray)
    e.g 480,640
    :return: RGB form image e.g 480,640,3. no real colors added - just shape as RGB
    see gray_to_BGR_and_back_test()
    """
    BGR_image = cv2.cvtColor(gray_img, cv2.COLOR_GRAY2BGR)
    return BGR_image


def BGR_img_to_gray(bgr_img: np.array) -> np.array:
    """
    :param bgr_img: from shape (x,y,3) - 3 channels
    e.g 480,640,3
    :return: gray image e.g 480,640. colors are replaced to gray colors
    see gray_to_BGR_and_back_test()
    """
    gray = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2GRAY)
    return gray


def BGR_img_to_RGB(bgr_img: np.array) -> np.array:
    """
    :param bgr_img: from shape (x,y,3) - 3 channels
    e.g 480,640,3
    :return: rgb image
    see BGR_img_to_RGB_and_back_test()
    """
    rgb_img = cv2.cvtColor(bgr_img, cv2.COLOR_BGR2RGB)
    return rgb_img


def RGB_img_to_BGR(rgb_img: np.array) -> np.array:
    """
    :param rgb_img: from shape (x,y,3) - 3 channels
    e.g 480,640,3
    :return: bgr image
    see BGR_img_to_RGB_and_back_test()
    """
    bgr_img = cv2.cvtColor(rgb_img, cv2.COLOR_RGB2BGR)
    return bgr_img


class CameraWu:
    def __init__(self, port: int, type_cam: str = 'cv2'):
        self.port = port
        self.type_cam = type_cam
        try:
            pip_err = 'Error {}. pip install {}'
            failed_err = 'Failed to open CameraWu({}) on port {}'.format(type_cam, port)
            if type_cam == 'acapture':
                try:
                    # noinspection PyPackageRequirements
                    import acapture  # pip install acapture
                    if mt.is_linux():  # NOT TESTED
                        acapture.camera_info()
                    if not self.check_port_valid(port):
                        self.cam = None
                        raise Exception(failed_err)
                    self.cam = acapture.open(port)
                    success, frame = self.cam.read()
                    if frame is None:
                        self.cam = None
                        raise Exception(failed_err)
                except ModuleNotFoundError as e:
                    self.cam = None
                    raise ModuleNotFoundError(pip_err.format(e, 'acapture'))
            elif type_cam == 'imutils':
                try:
                    # noinspection PyPackageRequirements,PyUnresolvedReferences
                    from imutils import video  # pip install imutils
                    self.cam = video.VideoStream(src=port).start()
                    frame = self.cam.read()
                    if frame is None:
                        self.cam = None
                        raise Exception(failed_err)
                except ModuleNotFoundError as e:
                    self.cam = None
                    raise ModuleNotFoundError(pip_err.format(e, 'imutils'))
            else:  # type_cam == 'cv2'
                self.cam = cv2.VideoCapture(port)
                if not self.cam.isOpened():
                    self.cam = None
                    raise Exception(failed_err)
        except Exception as e:
            raise e
        return

    @classmethod
    def open_camera(cls, port: int, type_cam: str = 'cv2'):
        try:
            cam = cls(port, type_cam)
            print('\tCameraWu({}) successfully open on port {}'.format(type_cam, port))
        except (ModuleNotFoundError, Exception) as e:
            mt.exception_error(e, real_exception=True, tabs=1)
            cam = None
        return cam

    @staticmethod
    def check_port_valid(port: int) -> bool:
        """
        :param port:
        :return:
        """
        temp_cam = cv2.VideoCapture(port)
        if temp_cam.isOpened():
            ret = True
            temp_cam.release()
        else:
            ret = False
        return ret

    def __del__(self):
        try:
            if self.cam is not None:
                if self.type_cam == 'acapture':
                    # noinspection PyUnresolvedReferences
                    self.cam.destroy()
                    mt.sleep(2, ack=True, tabs=2)  # acapture need a moment to release
                elif self.type_cam == 'imutils':
                    # noinspection PyUnresolvedReferences
                    self.cam.stop()
                    mt.sleep(2, ack=True, tabs=2)  # imutils need a moment to release
                else:  # type_cam == 'cv2'
                    if self.cam.isOpened():
                        self.cam.release()
                        mt.sleep(2, ack=True, tabs=2)  # just in case
                print('\tCameraWu({}) closed on port {}'.format(self.type_cam, self.port))
        except AttributeError as e:
            mt.exception_error('e {}. can\'t close CameraWu'.format(e), real_exception=True)
        return

    def read_img(self) -> (bool, np.array):
        try:
            frame = None
            if self.type_cam == 'acapture':
                _, frame = self.cam.read()
                if frame is not None:
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            elif self.type_cam == 'imutils':
                frame = self.cam.read()
            else:  # type_cam == 'cv2'
                if self.cam.isOpened():
                    # self.cam.release()
                    # self.cam = cv2.VideoCapture(self.port)
                    _, frame = self.cam.read()
                    # self.cam.set(cv2.CAP_PROP_POS_FRAMES, 0)
            success = False if frame is None else True
        except Exception as e:
            mt.exception_error('CameraWu: failed capture image. e {}'.format(e), real_exception=True)
            success, frame = False, None
        return success, frame


def add_text(
        cv_img: np.array,
        header: str,
        pos: tuple,
        text_color: str = 'white',
        with_rect: bool = True,
        bg_color: str = 'black',
        bg_font_scale: int = 1
):
    """
    :param cv_img:
    :param header:
    :param pos:
    :param text_color: str
    :param with_rect:
    :param bg_color: str
    :param bg_font_scale: int. 1 or 2 - pos changes by scale and font
    :return:
    see add_text_test()
    """
    x, y = pos
    if bg_font_scale == 1:  # big font and scale
        font: int = cv2.FONT_HERSHEY_DUPLEX
        font_scale: float = 1
        font_thickness: int = 1
        text_size, _ = cv2.getTextSize(header, font, font_scale, font_thickness)
        text_w, text_h = text_size
        rect_pt1 = (x, y - 20)
        rect_pt2 = (x + text_w, y + text_h - 20)
    else:
        font: int = cv2.FONT_HERSHEY_DUPLEX
        font_scale: float = 0.5
        font_thickness: int = 1
        text_size, _ = cv2.getTextSize(header, font, font_scale, font_thickness)
        text_w, text_h = text_size
        rect_pt1 = (x, y)
        rect_pt2 = (x + text_w, y + text_h - 20)
    text_color = pyplt.get_BGR_color(text_color)

    # print(pos, (x + text_w, y + text_h))
    if with_rect:
        bg_color = pyplt.get_BGR_color(bg_color)
        cv2.rectangle(cv_img, pt1=rect_pt1, pt2=rect_pt2, color=bg_color, thickness=-1)
    cv2.putText(cv_img, text=header, org=pos, fontFace=font, fontScale=font_scale, color=text_color,
                thickness=font_thickness)
    return


def add_header(
        cv_img: np.array,
        header: str,
        loc: str = pyplt.Location.TOP_LEFT.value,
        text_color: str = 'white',
        with_rect: bool = True,
        bg_color: str = 'black',
        bg_font_scale: int = 1
) -> None:
    """
    :param cv_img: frame
    :param header: some text. e.g. iteration, timestamp ....
    :param loc: supports pyplt.Location.TOP_LEFT.value,  pyplt.Location.BOTTOM_LEFT.value
    :param text_color: as string e.g. 'red'
    :param with_rect: add rect around header
    :param bg_color: background rect color
    :param bg_font_scale: int. 1 or 2 - pos changes by scale and font
    :return:
    see add_header_test()
    """
    if bg_font_scale == 1:
        if loc == pyplt.Location.BOTTOM_LEFT.value:
            pos = (0, cv_img.shape[0] - 5)
        else:  # loc == pyplt.Location.TOP_LEFT.value:
            pos = (0, 20)
    else:
        if loc == pyplt.Location.BOTTOM_LEFT.value:
            pos = (0, cv_img.shape[0] - 3)
        else:  # loc == pyplt.Location.TOP_LEFT.value:
            pos = (0, 10)

    add_text(cv_img, header, pos=pos, text_color=text_color, with_rect=with_rect, bg_color=bg_color,
             bg_font_scale=bg_font_scale)
    return


def get_dims_from_cap(cap: cv2.VideoCapture) -> tuple:
    """
    :param cap:
    frames size as (int, int)
    :return:
    see Mp4_creator_test()
    """
    out_dims = None
    if cap.isOpened():
        orig_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))  # float `width`
        orig_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))  # float `height`
        out_dims = (orig_width, orig_height)
    return out_dims


def get_frames_from_cap(cap: cv2.VideoCapture) -> int:
    """
    :param cap:
    :return:
    see Mp4_creator_test()
    """
    video_total_frames = None
    if cap.isOpened():
        video_total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) - 1
    return video_total_frames


class Mp4_creator:
    def __init__(self, out_full_path: str, out_fps: float, out_dims: tuple, tabs: int = 1):
        """
        :param out_full_path:
        :param out_fps:
        :param out_dims:
        :param tabs:
        see Mp4_creator_test()
        """
        # fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        fourcc = cv2.VideoWriter_fourcc(c1='m', c2='p', c3='4', c4='v')

        self.out_full_path = out_full_path
        self.out_fps = out_fps
        self.out_dims = out_dims
        self.tabs = tabs
        self.frames_count = 0
        if not os.path.exists(os.path.dirname(out_full_path)):
            mt.exception_error('Cant create cv2.VideoWriter', real_exception=False, tabs=self.tabs)
            mt.exception_error(mt.NOT_FOUND.format(os.path.dirname(out_full_path)), real_exception=False,
                               tabs=self.tabs)
            self.result = None
        else:
            self.result = cv2.VideoWriter(
                filename=out_full_path,
                fourcc=fourcc,
                fps=out_fps,
                frameSize=out_dims
            )
        return

    def __del__(self):
        if self.result is not None:
            self.result.release()
            self.result = None
        return

    def __str__(self):
        string = '{}Mp4_creator\n'.format(self.tabs * '\t')
        string += '{}\ttarget file path: {}\n'.format(self.tabs * '\t', self.out_full_path)
        string += '{}\tfps: {}\n'.format(self.tabs * '\t', self.out_fps)
        string += '{}\tout_dims: {}'.format(self.tabs * '\t', self.out_dims)
        return string

    def add_frame(self, frame: np.array, ack: bool = False, tabs: int = 1) -> None:
        if self.result is not None:
            cur_out_dims = (frame.shape[1], frame.shape[0])
            if cur_out_dims == self.out_dims:
                self.result.write(frame)
                self.frames_count += 1
                if ack:
                    print('{}frame {} added'.format(tabs * '\t', self.frames_count))
            else:
                mt.exception_error(
                    'Shapes mismatch: Frame.shape={}, Video shape={}'.format(cur_out_dims, self.out_dims),
                    real_exception=False, tabs=tabs)
        else:
            mt.exception_error('Cant write frame', real_exception=False, tabs=self.tabs)
        return

    def finalize(self):
        if self.result is not None:
            self.result.release()
            self.result = None
            print('{}File ready({} frames): {}'.format(self.tabs * '\t', self.frames_count, self.out_full_path))
        return



"""
:requires: torch, torchvision
for torch:
cpu: pip install torch
cuda v9 and v9.1:
    pip install torch==1.1.0 torchvision==0.3.0 -f https://download.pytorch.org/whl/cu90/torch_stable.html
"""

try:
    from wizzi_utils.torch.torch_tools import *
except ModuleNotFoundError as e:
    pass

from wizzi_utils.torch import test



import numpy as np
import abc
from tflite_runtime.interpreter import Interpreter
# noinspection PyPackageRequirements
import cv2
from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.pyplot import pyplot_tools as pyplt
from wizzi_utils.models import BaseModel
from wizzi_utils.models import models_configs as cfg


def get_pose_detection_models(ack: bool = False, tabs: int = 1) -> list:
    model_names = []
    count = 0
    for i, (m_name, m_dict) in enumerate(cfg.MODELS_CONFIG.items()):
        if m_dict['job'] == cfg.Jobs.POSE_DETECTION.value:
            if ack:
                count += 1
                mt.dict_as_table(table=m_dict, title='{}){}'.format(count, m_name), tabs=tabs)
            model_names.append(m_name)
    return model_names


class PdBaseModel(BaseModel):
    DEFAULT_COLOR_DICT = {
        'joint_c': 'red',  # joint color
        'edge_c': 'lightgreen',  # line color
        'text_c': 'white',  # label index color
    }
    NOT_FOUND_VALUE = -1
    NOT_FOUND_PAIR = [NOT_FOUND_VALUE, NOT_FOUND_VALUE]

    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            allowed_joint_names: list = None,
    ):
        super().__init__(save_load_dir=save_load_dir, model_name=model_name)
        if not self.model_name_valid(model_name):
            exit(-1)
        self.model_cfg = cfg.MODELS_CONFIG[model_name]
        self.joint_names = self.model_cfg['joint_names']
        self.pairs_indices = self.model_cfg['pairs_indices']
        self.allowed_joint_names = list(
            self.joint_names.values()) if allowed_joint_names is None else allowed_joint_names
        return

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def to_string(self, tabs: int = 1) -> str:
        print('abs method - needs implementation')
        exit(-1)
        return ''

    # noinspection PyUnusedLocal
    @abc.abstractmethod
    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        print('abs method - needs implementation')
        exit(-1)
        return []

    def draw_detections(
            self,
            detections: list,
            cv_img: np.array,
            colors_d: dict,
            draw_labels: bool,
    ) -> None:
        """
        :param detections: output of self.classify_cv_img()
        :param colors_d: colors in str form:
            bbox color
            label_bbox color
            text color
            e.g. colors_d={
                    'joint_c': 'red',
                    'edge_c': 'lightgreen',
                    'text_c': 'white',
                },
        :param cv_img: the same that was given input to self.classify_cv_img()
        :param draw_labels: draw part index
        :return:
        """
        if colors_d is None:
            colors_d = PdBaseModel.DEFAULT_COLOR_DICT
        edge_c = pyplt.get_BGR_color(colors_d['edge_c'])
        joint_c = pyplt.get_BGR_color(colors_d['joint_c'])
        text_c = colors_d['text_c']
        for detection in detections:  # each detection is a pose - currently supports just 1
            names = detection['joint_names_list']
            ids = detection['joint_ids_list']
            xys = detection['joint_x_y_list']
            # zs = detection['joint_z_list'] if 'joint_z_list' in detection else [self.NOT_FOUND_VALUE] * len(names)
            scores = detection['score_percentages_list']

            # DRAW LINES: lines connecting the edges
            for pair in self.pairs_indices:
                partA, partB = pair
                if xys[partA] != self.NOT_FOUND_PAIR and xys[partB] != self.NOT_FOUND_PAIR:
                    cv2.line(cv_img, pt1=tuple(xys[partA]), pt2=tuple(xys[partB]), color=edge_c, thickness=2)

            # DRAW JOINTS: circle and joint id if asked
            for name, jid, xy, score in zip(names, ids, xys, scores):
                if xy != self.NOT_FOUND_PAIR:
                    cv2.circle(cv_img, center=tuple(xy), radius=3, color=joint_c, thickness=-1)
                    if draw_labels:
                        label_xy = (xy[0] - 5, xy[1])
                        label = '{}'.format(jid)
                        cvt.add_text(cv_img, header=label, pos=label_xy, text_color=text_c, with_rect=False,
                                     bg_font_scale=2)
        return


class TflPdModel(PdBaseModel):
    def __init__(
            self,
            save_load_dir: str,
            model_name: str,
            threshold: float = None,
            allowed_joint_names: list = None
    ):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name in MODEL_CONF.keys()
        :param threshold: only detection above this threshold will be pass first filter
        :param allowed_joint_names: joint_names to track from the joint_names of the model
        example:
        see:
        """
        super().__init__(save_load_dir=save_load_dir, model_name=model_name, allowed_joint_names=allowed_joint_names)
        if not self.model_type_valid(self.model_cfg['model_type'], cfg.ModelType.PdTflNormal.value):
            exit(-1)
        if threshold is not None:
            self.model_cfg['threshold'] = threshold
        self.pairs_indices = self.model_cfg['pairs_indices']

        model_fp = "{}/{}.tflite".format(self.local_path, self.model_name)
        self._download_if_needed(local_path=model_fp, url_dict=self.model_cfg['tflite'])
        self.model_size = mt.file_or_folder_size(model_fp)
        # print('Loading {}(size {}, {} classes)'.format(self.model_name, self.model_size, len(self.labels)))
        self.interpreter = Interpreter(model_path=model_fp, num_threads=4)
        # allocate input output placeholders
        self.interpreter.allocate_tensors()
        self.input_details = self.interpreter.get_input_details()
        self.output_details = self.interpreter.get_output_details()
        _, input_height, input_width, _ = self.interpreter.get_input_details()[0]['shape']
        self.model_cfg['in_dims'] = (input_width, input_height)
        self.model_cfg['input_type'] = self.interpreter.get_input_details()[0]['dtype']
        quantization = self.interpreter.get_input_details()[0]['quantization']
        # need to normalize if model doesn't do quantization (then quantization == (0.0, 0))
        # else e.g. quantization == (0.0078125, 128) - no need to normalize
        self.model_cfg['normalize_RGB'] = True if quantization == (0.0, 0) else False

        # # you can print this to get more details on the model
        # mt.dict_as_table(self.interpreter.get_input_details()[0], title='input')
        # mt.dict_as_table(self.interpreter.get_output_details()[0], title='output')
        # mt.dict_as_table(self.interpreter.get_tensor_details()[0], title='tensor')
        return

    def to_string(self, tabs: int = 1) -> str:
        tabs_s = tabs * '\t'
        string = '{}{}'.format(tabs_s, mt.add_color(string='TflPdModel:', ops='underlined'))
        string += '\n\t{}name= {} (size {})'.format(tabs_s, self.model_name, self.model_size)
        string += '\n\t{}local_path= {}'.format(tabs_s, self.local_path)
        string += '\n\t{}{}'.format(tabs_s, mt.to_str(self.allowed_joint_names, 'allowed_classes'))
        string += '\n{}'.format(mt.dict_as_table(self.model_cfg, title='cfg', fp=6, ack=False, tabs=tabs + 1))
        return string

    def prepare_input(self, cv_img: np.array) -> np.array:
        """
        :param cv_img:
        resize and change dtype to predefined params
        :return:
        """
        img_RGB = cvt.BGR_img_to_RGB(cv_img)

        if self.model_cfg['normalize_RGB']:
            center = 127.5
            img_RGB = (img_RGB / center) - 1  # normalize image
        img = cv2.resize(img_RGB, self.model_cfg['in_dims'])  # size of this model input
        img_processed = np.expand_dims(img, axis=0).astype(self.model_cfg['input_type'])  # a,b,c -> 1,a,b,c
        return img_processed

    def run_network(self, img_preprocessed: np.array) -> None:
        self.interpreter.set_tensor(self.input_details[0]['index'], img_preprocessed)  # set input tensor
        self.interpreter.invoke()  # run
        return

    def extract_results(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: cv image
        :param fp: float precision on the score percentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts:
        dict is a detection of pose
            each has items:
                'joint_names_list':  - all joint names like in the config
                'joint_ids_list': - all joint ids like in the config
                'joint_x_y_list': - if joint found: it's x,y values
                'joint_z_list': - if joint found: it's z value
                'score_percentages_list': - if joint found: it's confidence in 0-100%,
        """
        # get results
        depth = 5  # each points has x,y,z,visibility,presence
        # full pose -> 195/5=39. but i think there are only 33
        outputs = self.interpreter.get_tensor(self.output_details[0]['index'])[0]  # points numpy

        if ack:
            title_suffix = '' if img_title is None else '{} '.format(img_title)
            title = '{} detection on image {}{}:'.format(self.model_name, title_suffix, cv_img.shape)
            print('{}{}'.format(tabs * '\t', title))
            print('{}Meta_data:'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(outputs, 'outputs')))
            print('{}Detections:'.format(tabs * '\t'))

        detections = []  # each detection is a set of joint
        img_h, img_w = cv_img.shape[0], cv_img.shape[1]
        net_w, net_h = self.model_cfg['in_dims']
        for pose_id in range(1):  # currently supports 1 pose
            detection_d = {
                'joint_names_list': [],
                'joint_ids_list': [],
                'joint_x_y_list': [],
                'joint_z_list': [],
                'score_percentages_list': [],
            }
            for joint_id, joint_name in self.joint_names.items():
                detection_d['joint_names_list'].append(joint_name)
                detection_d['joint_ids_list'].append(joint_id)
                x = int(outputs[joint_id * depth + 0] * img_w / net_w)
                y = int(outputs[joint_id * depth + 1] * img_h / net_h)
                z = int(outputs[joint_id * depth + 2])

                visibility = outputs[joint_id * depth + 3]
                visibility = 1 / (1 + np.exp(visibility))  # reverse sigmoid
                presence = outputs[joint_id * depth + 4]
                presence = 1 / (1 + np.exp(presence))  # reverse sigmoid
                score_frac = 1 - max(visibility, presence)  # change from err to acc: acc = 1-err
                score_percentage = round(score_frac * 100, fp)

                if self.model_cfg['threshold'] <= score_frac <= 1.0 and joint_name in self.allowed_joint_names:
                    detection_d['joint_x_y_list'].append([x, y])
                    detection_d['joint_z_list'].append(z)
                    detection_d['score_percentages_list'].append(score_percentage)
                else:
                    detection_d['joint_x_y_list'].append(self.NOT_FOUND_PAIR)
                    detection_d['joint_z_list'].append(self.NOT_FOUND_VALUE)
                    detection_d['score_percentages_list'].append(self.NOT_FOUND_VALUE)

                if ack:
                    d_msg = '{}\tpose {}: {}({}): xy={}, z={}, score=({}%)'
                    print(d_msg.format(tabs * '\t', pose_id, joint_name, joint_id, detection_d['joint_x_y_list'][-1],
                                       detection_d['joint_z_list'][-1], detection_d['score_percentages_list'][-1]))
            detections.append(detection_d)
        return detections

    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """

        img_preprocessed = self.prepare_input(cv_img)
        self.run_network(img_preprocessed)
        detections = self.extract_results(
            cv_img=cv_img,
            fp=fp,
            ack=ack,
            tabs=tabs,
            img_title=img_title
        )
        return detections


class Cv2PdModel(PdBaseModel):
    BAD_P = (-1, -1)

    def __init__(self,
                 save_load_dir: str,
                 model_name: str,
                 allowed_joint_names: list = None,
                 threshold: float = None,
                 in_dims: tuple = None,
                 scalefactor: float = None,
                 mean: tuple = None,
                 swapRB: bool = None,
                 crop: bool = None,
                 ):
        """
        :param save_load_dir: where the model is saved (or will be if not exists)
        :param model_name: valid name in MODEL_CONF.keys()
        for all the following - if is None: take default value from MODELS_DNN_OBJECT_DETECTION_META_DATA['model_name']
        :param threshold: only detection above this threshold will be returned
        :param in_dims:
        :param scalefactor:
        :param mean:
        :param swapRB:
        :param crop:
        example:
        see:
        """
        super().__init__(save_load_dir=save_load_dir, model_name=model_name, allowed_joint_names=allowed_joint_names)
        if not self.model_type_valid(self.model_cfg['model_type'], cfg.ModelType.PdCvNormal.value):
            exit(-1)
        self.pose_net = None
        if self.model_cfg['family'] == cfg.DnnFamily.Caffe.value:
            model_prototxt = "{}/{}.prototxt".format(self.local_path, self.model_name)
            model_caffe = "{}/{}.caffemodel".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_prototxt, url_dict=self.model_cfg['prototxt'])
            self._download_if_needed(local_path=model_caffe, url_dict=self.model_cfg['caffemodel'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_prototxt), mt.file_or_folder_size(model_caffe)]
            self.pose_net = cv2.dnn.readNetFromCaffe(prototxt=model_prototxt, caffeModel=model_caffe)

        elif self.model_cfg['family'] == cfg.DnnFamily.Darknet.value:
            model_cfg = "{}/{}.cfg".format(self.local_path, self.model_name)
            model_weights = "{}/{}.weights".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_cfg, url_dict=self.model_cfg['cfg'])
            self._download_if_needed(local_path=model_weights, url_dict=self.model_cfg['weights'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_cfg), mt.file_or_folder_size(model_weights)]
            self.pose_net = cv2.dnn.readNetFromDarknet(cfgFile=model_cfg, darknetModel=model_weights)

        elif self.model_cfg['family'] == cfg.DnnFamily.TF.value:
            model_pbtxt = "{}/{}.pbtxt".format(self.local_path, self.model_name)
            model_pb = "{}/{}.pb".format(self.local_path, self.model_name)
            self._download_if_needed(local_path=model_pbtxt, url_dict=self.model_cfg['pbtxt'])
            self._download_if_needed(local_path=model_pb, url_dict=self.model_cfg['pb'])
            self.model_resources_sizes = [mt.file_or_folder_size(model_pbtxt), mt.file_or_folder_size(model_pb)]
            self.pose_net = cv2.dnn.readNetFromTensorflow(model=model_pb, config=model_pbtxt)

        if self.pose_net is None:
            mt.exception_error('Failed to create network', real_exception=False)
            exit(-1)

        self.pairs_indices = self.model_cfg['pairs_indices']

        if threshold is not None:
            self.model_cfg['threshold'] = threshold
        if in_dims is not None:
            self.model_cfg['in_dims'] = in_dims
        if scalefactor is not None:
            self.model_cfg['scalefactor'] = scalefactor
        if mean is not None:
            self.model_cfg['mean'] = mean
        if swapRB is not None:
            self.model_cfg['swapRB'] = swapRB
        if crop is not None:
            self.model_cfg['crop'] = crop

        return

    def to_string(self, tabs: int = 1) -> str:
        tabs_s = tabs * '\t'
        string = '{}{}'.format(tabs_s, mt.add_color(string='Cv2PdModel:', ops='underlined'))
        string += '\n\t{}name= {} (resources size: {})'.format(tabs_s, self.model_name, self.model_resources_sizes)
        string += '\n\t{}local_path={}'.format(tabs_s, self.local_path)
        string += '\n\t{}{}'.format(tabs_s, mt.to_str(self.allowed_joint_names, 'allowed_joint_names'))
        string += '\n{}'.format(mt.dict_as_table(self.model_cfg, title='conf', fp=6, ack=False, tabs=tabs + 1))
        return string

    def detect_cv_img(
            self,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param cv_img: open cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts. see extract_results()
        """
        inpBlob = cv2.dnn.blobFromImage(
            image=cv_img,
            scalefactor=self.model_cfg['scalefactor'],
            size=self.model_cfg['in_dims'],
            mean=self.model_cfg['mean'],
            swapRB=self.model_cfg['swapRB'],
            crop=self.model_cfg['crop'])
        self.pose_net.setInput(inpBlob)
        outputs = self.pose_net.forward()
        # print('{}\t{}'.format(tabs * '\t', mt.to_str(outputs, 'outputs')))

        if len(outputs) > 0:
            outputs = outputs[0]

        detections = self.extract_results(
            outputs=outputs,
            cv_img=cv_img,
            fp=fp,
            ack=ack,
            tabs=tabs,
            img_title=img_title
        )
        return detections

    def extract_results(
            self,
            outputs: np.array,
            cv_img: np.array,
            fp: int = 2,
            ack: bool = False,
            tabs: int = 1,
            img_title: str = None
    ) -> list:
        """
        :param outputs: np array that contains the outputs for the cv_img
        :param cv_img: cv image
        :param fp: float precision on the score precentage. e.g. fp=2: 0.1231231352353 -> 12.31%
        :param ack: if True: print meta data and detections
        :param tabs: if ack True: tabs for print
        :param img_title: if ack True: title for print
        :return: list of dicts:
        dict is a detection of pose
            each has items:
                'joint_names_list':  - all joint names like in the config
                'joint_ids_list': - all joint ids like in the config
                'joint_x_y_list': - if joint found: it's x,y values
                'joint_z_list': - if joint found: it's z value
                'score_percentages_list': - if joint found: it's confidence in 0-100%,
        """

        if ack:
            title_suffix = '' if img_title is None else '{} '.format(img_title)
            title = '{} detection on image {}{}:'.format(self.model_name, title_suffix, cv_img.shape)
            print('{}{}'.format(tabs * '\t', title))
            print('{}Meta_data:'.format(tabs * '\t'))
            print('{}\t{}'.format(tabs * '\t', mt.to_str(outputs, 'outputs')))
            print('{}Detections:'.format(tabs * '\t'))

        detections = []  # each detection is a set of joint
        img_h, img_w = cv_img.shape[0], cv_img.shape[1]
        net_h, net_w = outputs.shape[1], outputs.shape[2]

        for pose_id in range(1):  # currently supports 1 pose
            detection_d = {
                'joint_names_list': [],
                'joint_ids_list': [],
                'joint_x_y_list': [],
                'score_percentages_list': [],
            }

            for joint_id, joint_name in self.joint_names.items():
                detection_d['joint_names_list'].append(joint_name)
                detection_d['joint_ids_list'].append(joint_id)
                probMap = outputs[joint_id, :, :]  # confidence map.
                minVal, score_frac, minLoc, point = cv2.minMaxLoc(probMap)  # Find global maxima of the probMap.
                # Scale the point to fit on the original image
                x = int((img_w * point[0]) / net_w)
                y = int((img_h * point[1]) / net_h)
                score_percentage = round(score_frac * 100, fp)

                if self.model_cfg['threshold'] <= score_frac <= 1.0 and joint_name in self.allowed_joint_names:
                    detection_d['joint_x_y_list'].append([x, y])
                    detection_d['score_percentages_list'].append(score_percentage)
                else:
                    detection_d['joint_x_y_list'].append(self.NOT_FOUND_PAIR)
                    detection_d['score_percentages_list'].append(self.NOT_FOUND_VALUE)

                if ack:
                    d_msg = '{}\tpose {}: {}({}): xy={}, score=({}%)'
                    print(d_msg.format(tabs * '\t', pose_id, joint_name, joint_id, detection_d['joint_x_y_list'][-1],
                                       detection_d['score_percentages_list'][-1]))
            detections.append(detection_d)
        return detections



"""
:requires: torch for compare_images_multi_sets_squeezed_test()
:sub_package: open_cv with the test
"""
try:
    from wizzi_utils.pyplot.test.test_pyplot_tools import *
except ModuleNotFoundError as e:
    pass



from wizzi_utils.misc import misc_tools as mt
from wizzi_utils.misc.test import test_misc_tools as mtt
from wizzi_utils.open_cv import open_cv_tools as cvt
from wizzi_utils.open_cv.test import test_open_cv_tools as cvtt
from wizzi_utils.pyplot import pyplot_tools as pyplt
from wizzi_utils.models import models_configs as cfg
from wizzi_utils.models import BaseModel
from wizzi_utils.models import object_detection as od
from wizzi_utils.models import pose_detection as pd
import numpy as np
import os
# noinspection PyPackageRequirements
import cv2

IMAGE_MS = 5000
CAM_FRAMES = 3
VID_X_FRAMES = 80


def _get_models(model_names: list) -> list:
    models = []
    for model_name in model_names:
        if BaseModel.model_name_valid(model_name):
            model_cfg = cfg.MODELS_CONFIG[model_name]
            m_save_dir = '{}/{}/{}'.format(mtt.MODELS, cfg.MODELS_CONFIG[model_name]['job'], model_name)
            if model_cfg['model_type'] == cfg.ModelType.OdCvNormal.value:
                model = od.Cv2OdModel(
                    save_load_dir=m_save_dir,
                    model_name=model_name,
                )
            elif model_cfg['model_type'] == cfg.ModelType.OdTflNormal.value:
                model = od.TflOdModel(
                    save_load_dir=m_save_dir,
                    model_name=model_name,
                )
            elif model_cfg['model_type'] == cfg.ModelType.PdCvNormal.value:
                model = pd.Cv2PdModel(
                    save_load_dir=m_save_dir,
                    model_name=model_name,
                )
            elif model_cfg['model_type'] == cfg.ModelType.PdTflNormal.value:
                model = pd.TflPdModel(
                    save_load_dir=m_save_dir,
                    model_name=model_name,
                    # allowed_joint_names=['nose', 'leftEyeInside', 'leftEye']
                )
            else:
                model = None
                mt.exception_error('model type not found')
                exit(-1)
            print(model.to_string(tabs=1))
            models.append(model)
    return models


def od_run(
        model: (od.TflOdModel, od.Cv2OdModel),
        cv_img: np.array,
        fps: mt.FPS = None,
        with_tf: bool = False,
        with_sub: bool = False
):
    if fps is not None:
        fps.start()
    detections = model.detect_cv_img(cv_img, ack=False)
    if fps is not None:
        fps.update(ack_progress=True, tabs=1, with_title=True)
    if with_tf:
        detections = model.add_traffic_light_to_detections(
            detections,
            traffic_light_p={
                'up': 0.2,
                'mid': 0.3,
                'down': 0.4
            }
        )
    if with_sub:
        detections = model.add_sub_sub_image_to_detection(
            detections,
            cv_img=cv_img,
            bbox_image_p={
                'x_start': 0.2,
                'x_end': 0.8,
                'y_start': 1,
                'y_end': 0.5,
            },
        )
    model.draw_detections(
        detections,
        colors_d={
            'bbox': 'r',
            'label_bbox': 'black',
            'text': 'white',
            'sub_image': 'blue',
            'person_bbox': 'lightgreen',
            'dog_bbox': 'blue',
        },
        cv_img=cv_img,
        draw_labels=True,
    )
    return cv_img


def pd_run(
        model: (pd.TflPdModel, pd.Cv2PdModel),
        cv_img: np.array,
        fps: mt.FPS = None,
):
    if fps is not None:
        fps.start()
    detections = model.detect_cv_img(cv_img, ack=False)
    if fps is not None:
        fps.update(ack_progress=True, tabs=1, with_title=True)
    model.draw_detections(
        detections,
        colors_d={
            'joint_c': 'red',
            'edge_c': 'lightgreen',
            'text_c': 'black',
        },
        cv_img=cv_img,
        draw_labels=True,
    )
    return cv_img


def od_or_pd_Model_image_test(
        cv_img: np.array,
        model_names: list,
        ms: int = cvtt.BLOCK_MS_NORMAL,
        dis_size: tuple = (640, 480),
        grid: tuple = (1, 1),
):
    mt.get_function_name(ack=True, tabs=0)
    orig_shape = '{}'.format(cv_img.shape)
    cv_imgs_post = []
    models = _get_models(model_names)
    for model in models:
        cv_img_clone = cv_img.copy()
        fps = mt.FPS(summary_title='{}'.format(model.model_name))
        if isinstance(model, (od.Cv2OdModel, od.TflOdModel)):
            cv_img_post = od_run(model, cv_img_clone, fps=fps, with_tf=False, with_sub=False)
        elif isinstance(model, (pd.Cv2PdModel, pd.TflPdModel)):
            cv_img_post = pd_run(model, cv_img_clone, fps=fps)
        else:
            cv_img_post = None

        cvt.add_header(cv_img_post, header=model.model_name, loc=pyplt.Location.BOTTOM_LEFT.value, bg_font_scale=1)
        cv_img_post = cv2.resize(cv_img_post, dis_size, interpolation=cv2.INTER_AREA)
        cv_imgs_post.append(cv_img_post)
    cvt.display_open_cv_images(
        cv_imgs_post,
        ms=ms,
        title=orig_shape,
        loc=pyplt.Location.CENTER_CENTER.value,
        grid=grid,
        resize=None,
        header=None,
        save_path=None
    )
    cv2.destroyAllWindows()
    return


def _od_or_pd_Model_cam_test(
        cam: (cv2.VideoCapture, cvt.CameraWu),
        model_names: list = None,
        max_frames: int = CAM_FRAMES,
        work_every_x_frames: int = 1,
        dis_size: tuple = (640, 480),
        grid: tuple = (1, 1),
):
    if len(model_names) == 1:
        mp4_out_dir = '{}/{}/{}'.format(mtt.VIDEOS_OUTPUTS, mt.get_function_name(), model_names[0])
        if not os.path.exists(mp4_out_dir):
            mt.create_dir(mp4_out_dir)
        out_fp = '{}/{}_detected.mp4'.format(mp4_out_dir, 'cam')
        out_dims = (dis_size[0] * grid[1], dis_size[1] * grid[0])
        mp4 = cvt.Mp4_creator(
            out_full_path=out_fp,
            out_fps=20.0,
            out_dims=out_dims
        )
        print(mp4)
    else:
        mp4 = None
    models = _get_models(model_names)
    fps_list = [mt.FPS(summary_title='{}'.format(model.model_name)) for model in models]
    for i in range(max_frames):
        if isinstance(cam, cv2.VideoCapture):
            success, cv_img = cam.read()
        else:
            success, cv_img = cam.read_img()
        if i % work_every_x_frames != 0:  # s
            # do only x frames
            continue
        cv_imgs_post = []
        if success:
            for model, fps in zip(models, fps_list):
                cv_img_clone = cv_img.copy()
                if isinstance(model, (od.Cv2OdModel, od.TflOdModel)):
                    cv_img_post = od_run(model, cv_img_clone, fps=fps, with_tf=False, with_sub=False)
                elif isinstance(model, (pd.Cv2PdModel, pd.TflPdModel)):
                    cv_img_post = pd_run(model, cv_img_clone, fps=fps)
                else:
                    cv_img_post = None
                cvt.add_header(cv_img_clone, header=model.model_name, loc=pyplt.Location.BOTTOM_LEFT.value,
                               bg_font_scale=1)
                cv_img_post = cv2.resize(cv_img_post, dis_size, interpolation=cv2.INTER_AREA)
                cv_imgs_post.append(cv_img_post)
            if mp4 is not None:
                mp4.add_frame(cv_imgs_post[0])
        cvt.display_open_cv_images(
            cv_imgs_post,
            ms=1,
            title='cam 0',
            loc=pyplt.Location.CENTER_CENTER.value,
            grid=grid,
            resize=None,
            header='{}/{}'.format(i + 1, max_frames),
            save_path=None
        )
    for fps in fps_list:
        fps.finalize()
    if mp4 is not None:
        mp4.finalize()
    cv2.destroyAllWindows()
    return


def od_or_pd_Model_cam_test(
        model_names: list = None,
        max_frames: int = CAM_FRAMES,
        dis_size: tuple = (640, 480),
        grid: tuple = (1, 1),
):
    mt.get_function_name(ack=True, tabs=0)
    cam = cvt.CameraWu.open_camera(port=0, type_cam='cv2')
    if cam is not None:
        _od_or_pd_Model_cam_test(
            cam=cam,
            model_names=model_names,
            max_frames=max_frames,
            work_every_x_frames=1,
            dis_size=dis_size,
            grid=grid,
        )
    return


def od_or_pd_Model_video_test(
        model_names: list,
        vid_path: str,
        work_every_x_frames: int = 1,
        dis_size: tuple = (640, 480),
        grid: tuple = (1, 1),
):
    mt.get_function_name(ack=True, tabs=0)
    cam = cv2.VideoCapture(vid_path)
    if cam is not None:
        _od_or_pd_Model_cam_test(
            cam=cam,
            model_names=model_names,
            max_frames=cvt.get_frames_from_cap(cam),
            work_every_x_frames=work_every_x_frames,
            dis_size=dis_size,
            grid=grid,
        )
    return


def __get_od_dict(op: int = 1):
    if op == 1:
        od_solo = {
            'model_names': ['coco_ssd_mobilenet_v1_1_0_quant_2018_06_29'],
            'dis_size': (640, 480),
            'grid': (1, 1),
        }
        od_meta_dict = od_solo
    elif op == 2:
        od_dual = {
            'model_names': [
                'coco_ssd_mobilenet_v1_1_0_quant_2018_06_29',
                'yolov4_tiny'
            ],
            'dis_size': (320, 240),
            'grid': (1, 2),
        }
        od_meta_dict = od_dual
    elif op == 3:
        od_all_models = {
            'model_names': od.get_object_detection_models(ack=False),
            'dis_size': (160, 120),
            'grid': (5, 6),
        }
        od_meta_dict = od_all_models
    else:
        od_meta_dict = None
    return od_meta_dict


def __get_pd_dict(op: int = 1):
    if op == 1:
        pd_solo = {
            'model_names': ['pose_landmark_lite'],
            'dis_size': (640, 480),
            'grid': (1, 1),
        }
        pd_meta_dict = pd_solo
    elif op == 2:
        pd_dual = {
            'model_names': [
                'pose_landmark_lite',
                'openpose_pose_mpi'
            ],
            'dis_size': (640, 480),
            'grid': (1, 2),
        }
        pd_meta_dict = pd_dual
    elif op == 3:
        all_models_names = pd.get_pose_detection_models(ack=False)
        # remove hand pose - not the same work
        all_models_names.remove('hand_pose')
        pd_all_models = {
            'model_names': all_models_names,
            'dis_size': (480, 360),
            'grid': (2, 3),
        }
        pd_meta_dict = pd_all_models
    else:
        pd_meta_dict = None
    return pd_meta_dict


def test_all():
    print('{}{}:'.format('-' * 5, mt.get_base_file_and_function_name()))

    od_image_test = False
    od_cam_test = False
    od_vid_test = False
    pd_image_test = False
    pd_cam_test = False
    pd_vid_test = False

    # od.get_object_detection_models(ack=True)
    # pd.get_pose_detection_models(ack=True)
    # exit(-12)
    od_meta_dict = __get_od_dict(op=2)
    pd_meta_dict = __get_pd_dict(op=2)

    # OBJECT DETECTION TESTS
    if od_image_test:
        cv_img = cvtt.load_img_from_web(mtt.DOG, ack=False)
        od_or_pd_Model_image_test(
            cv_img=cv_img,
            model_names=od_meta_dict['model_names'],
            ms=IMAGE_MS,
            dis_size=od_meta_dict['dis_size'],
            grid=od_meta_dict['grid'],
        )
    if od_cam_test:
        od_or_pd_Model_cam_test(
            model_names=od_meta_dict['model_names'],
            max_frames=CAM_FRAMES,
            dis_size=od_meta_dict['dis_size'],
            grid=od_meta_dict['grid'],
        )
    if od_vid_test:
        vid_path = cvtt.get_vid_from_web(name=mtt.DOG1)
        od_or_pd_Model_video_test(
            model_names=od_meta_dict['model_names'],
            vid_path=vid_path,
            work_every_x_frames=VID_X_FRAMES,
            dis_size=od_meta_dict['dis_size'],
            grid=od_meta_dict['grid'],
        )
    # POSE DETECTION TESTS
    if pd_image_test:
        cv_img = cvtt.load_img_from_web(mtt.F_MODEL, ack=False)
        # cv_img = cvtt.load_img_from_web(mtt.HAND, ack=False)  # if testing 'hand_pose', use HAND image
        od_or_pd_Model_image_test(
            cv_img=cv_img,
            model_names=pd_meta_dict['model_names'],
            ms=IMAGE_MS,
            dis_size=pd_meta_dict['dis_size'],
            grid=pd_meta_dict['grid'],
        )
    if pd_cam_test:
        od_or_pd_Model_cam_test(
            model_names=pd_meta_dict['model_names'],
            max_frames=CAM_FRAMES,
            dis_size=pd_meta_dict['dis_size'],
            grid=pd_meta_dict['grid'],
        )
    if pd_vid_test:
        vid_path = cvtt.get_vid_from_web(name=mtt.DOG1)
        od_or_pd_Model_video_test(
            model_names=pd_meta_dict['model_names'],
            vid_path=vid_path,
            work_every_x_frames=VID_X_FRAMES,
            dis_size=pd_meta_dict['dis_size'],
            grid=pd_meta_dict['grid'],
        )

    print('{}'.format('-' * 20))
    return



from wizzi_utils import misc_tools as mt  # misc tools
# noinspection PyPackageRequirements
import tflite_runtime


# from tflite_runtime.interpreter import Interpreter

# try:
#     # noinspection PyUnresolvedReferences
#     interpreter_found = tflite_runtime.interpreter.Interpreter
#     print('success using tflite_runtime.interpreter.Interpreter')
# except AttributeError:
#     # print('failure using tflite_runtime.interpreter.Interpreter')
#     try:
#         # noinspection PyPackageRequirements,PyUnresolvedReferences
#         import tensorflow as tf
#
#         interpreter_found = tf.lite.Interpreter
#         print('success using tf.lite.Interpreter')
#     except (ImportError, AttributeError):
#         pass
#         print('failure using tf.lite.Interpreter')


def get_tflite_version(ack: bool = False, tabs: int = 1) -> str:
    """
    :param ack:
    :param tabs:
    :return:
    see get_tflite_version_test()
    """
    string = mt.add_color('{}* TFLite Version {}'.format(tabs * '\t', tflite_runtime.__version__), ops=mt.SUCCESS_C)
    # string += mt.add_color(' - GPU detected ? ', op1=mt.SUCCESS_C)
    # if gpu_detected():
    #     string += mt.add_color('True', op1=mt.SUCCESS_C2[0], extra_ops=mt.SUCCESS_C2[1])
    # else:
    #     string += mt.add_color('False', op1=mt.FAIL_C2[0], extra_ops=mt.FAIL_C2[1])
    if ack:
        print(string)
    return string


def gpu_detected() -> bool:
    """
    :return:
    TODO FUTURE - maybe check if threads available
    """
    return False
